from torch.optim.optimizer import Optimizer

from pydgn.experiment.util import s2c
from pydgn.static import *
from pydgn.training.event.handler import EventHandler
from pydgn.training.event.state import State


class Scheduler(EventHandler):
    """
    Scheduler is the main event handler for schedulers. Just pass a PyTorch
    scheduler together with its arguments in the configuration file.

    Args:
        scheduler_class_name (str): dotted path to class name of the scheduler
        optimizer (:class:`torch.optim.optimizer`): the Pytorch optimizer to
            use. **This is automatically recovered by PyDGN when
            providing an optimizer**
        kwargs: additional parameters for the specific scheduler to be used

    """

    def __init__(
        self, scheduler_class_name: str, optimizer: Optimizer, **kwargs: dict
    ):
        self.scheduler = s2c(scheduler_class_name)(optimizer, **kwargs)

    def on_fit_start(self, state: State):
        """
        Loads the scheduler state if already present in the state_dict
        of a checkpoint

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if self.scheduler and state.scheduler_state is not None:
            self.scheduler.load_state_dict(state.scheduler_state)

    def on_epoch_end(self, state: State):
        """
        Updates the scheduler state with the current one for checkpointing

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        state.update(scheduler_state=self.scheduler.state_dict())


class EpochScheduler(Scheduler):
    """
    Implements a scheduler which uses epochs to modify the step size
    """

    def on_training_epoch_end(self, state: State):
        """
        Performs a scheduler's step at the end of the training epoch.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        self.scheduler.step()


class MetricScheduler(Scheduler):
    """
    Implements a scheduler which uses variations in the metric of interest
    to modify the step size

    Args:
        scheduler_class_name (str): dotted path to class name of the scheduler
        use_loss (str): whether to monitor scores or losses
        monitor (str): the metric to monitor. The format is
            ``[TRAINING|VALIDATION]_[METRIC NAME]``, where
            ``TRAINING`` and ``VALIDATION`` are defined in ``pydgn.static``
        optimizer (:class:`torch.optim.optimizer`): the Pytorch optimizer
            to use. **This is automatically recovered by PyDGN when
            providing an optimizer**
        kwargs: additional parameters for the specific scheduler to be used
    """

    def __init__(
        self,
        scheduler_class_name: str,
        use_loss: bool,
        monitor: str,
        optimizer: Optimizer,
        **kwargs: dict,
    ):
        self.scheduler = s2c(scheduler_class_name)(optimizer, **kwargs)
        self.use_loss = use_loss
        self.monitor = monitor

    def on_epoch_end(self, state: State):
        """
        Updates the state of the scheduler according to a metric to monitor
        at each epoch. Finally, loads the scheduler state if already present
        in the state_dict of a checkpoint

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """

        monitor_main_key = LOSSES if self.use_loss else SCORES
        assert (
            self.monitor in state.epoch_results[monitor_main_key]
        ), f"{self.monitor} not found in epoch_results"
        metric = state.epoch_results[monitor_main_key][self.monitor]
        self.scheduler.step(metric)

        # Do not forget to store dict
        super().on_epoch_end(state)



# EVALUATOR - PROGRESS MANAGER
START_CONFIG = "START_CONFIG"
END_CONFIG = "END_CONFIG"
END_FINAL_RUN = "END_FINAL_RUN"
OUTER_FOLD = "outer_fold"
INNER_FOLD = "inner_fold"
CONFIG_ID = "config_id"
RUN_ID = "run_id"
ELAPSED = "elapsed"
EXP_NAME = "exp_name"
SEED = "seed"
GRID_SEARCH = "grid"
RANDOM_SEARCH = "random"
NUM_SAMPLES = "num_samples"
SAMPLE_METHOD = "sample_method"
DATASET = "dataset"
CONFIG = "config"
MODEL_ASSESSMENT = "MODEL_ASSESSMENT"
FOLDS = "folds"
TR_LOSS = "training_loss"
VL_LOSS = "validation_loss"
TR_SCORE = "training_score"
VL_SCORE = "validation_score"
BEST_CONFIG = "best_config"
OUTER_TRAIN = "outer_train"
OUTER_VALIDATION = "outer_validation"
OUTER_TEST = "outer_test"

# SPLITTER
DATA_SPLITTER_BASE_PATH = "pydgn.data.splitter."

# ENGINE
DEFAULT_ENGINE_CALLBACK = (
    "pydgn.training.callback.engine_callback.EngineCallback"
)

# STATE
TRAINING = "training"
EVALUATION = "evaluation"
VALIDATION = "validation"
TEST = "test"
MAIN_LOSS = "main_loss"
MAIN_SCORE = "main_score"
LOSSES = "losses"
LOSS = "loss"
SCORES = "scores"
SCORE = "score"
MODEL_STATE = "model_state"
OPTIMIZER_STATE = "optimizer_state"
SCHEDULER_STATE = "scheduler_state"
BEST_EPOCH = "best_epoch"
BEST_EPOCH_RESULTS = "best_epoch_results"
EPOCH = "epoch"
STOP_TRAINING = "stop_training"
LAST_CHECKPOINT_FILENAME = "last_checkpoint.pth"
BEST_CHECKPOINT_FILENAME = "best_checkpoint.pth"
BATCH_LOSS_EXTRA = "batch_loss_extra"

# STRING FORMATTING
TENSORBOARD = "tensorboard"
EMB_TUPLE_SUBSTR = "_embeddings_tuple"
ATOMIC_SAVE_EXTENSION = ".part"
CLASS_NAME = "class_name"
ARGS = "args"
MODEL_STATE_DICT = "model_state_dict"
OPTIMIZER_STATE_DICT = "optimizer_state_dict"
AVG = "avg"
STD = "std"

# SET UP
CUDA = "cuda"
PYDGN_RAY_NUM_GPUS_PER_TASK = "PYDGN_RAY_NUM_GPUS_PER_TASK"
MIN = "min"
MAX = "max"

# CLI
CONFIG_FILE_CLI_ARGUMENT = "--config-file"
CONFIG_FILE = "config_file"
DEBUG_CLI_ARGUMENT = "--debug"

DATA_ROOT = "data_root"
DATA_SPLITS_FILE = "data_splits_file"
DATASET_CLASS = "dataset_class"
DATASET_GETTER = "dataset_getter"
evaluate_every = "evaluate_every"
EVALUATE_EVERY = "evaluate_every"
DATA_LOADER = "data_loader"
DATA_LOADER_ARGS = "data_loader_args"
DATASET_NAME = "dataset_name"
DEBUG = "debug"
DEVICE = "device"
EXPERIMENT = "experiment"
FINAL_TRAINING_RUNS = "final_training_runs"
GPUS_PER_TASK = "gpus_per_task"
HIGHER_RESULTS_ARE_BETTER = "higher_results_are_better"
MAX_CPUS = "max_cpus"
MAX_GPUS = "max_gpus"
GPUS_SUBSET = "gpus_subset"
MODEL = "model"
RESULT_FOLDER = "result_folder"

# TELEGRAM BOT
TELEGRAM_CONFIG_FILE = "telegram_config_file"
TELEGRAM_BOT_TOKEN = "bot_token"
TELEGRAM_BOT_CHAT_ID = "bot_chat_ID"
LOG_MODEL_SELECTION = "log_model_selection"
LOG_FINAL_RUNS = "log_final_runs"



import bisect
import datetime
import time

from pydgn.training.event.handler import EventHandler


# Decorator class. Ideas taken from "Fluent Python" book
class Profiler:
    r"""
    A decorator class that is applied to a
    :class:`~pydgn.training.event.handler.EventHandler` object implementing a
    set of callback functions. For each callback, the Profiler stores the
    average and total running time across epochs.
    When the experiment terminates (either correctly or abruptly) the
    Profiler can produce a report to be stored in the experiment's log file.

    The Profiler is used as a singleton, and it produces wrappers that update
    its own state.

    Args:
        threshold (float): used to filter out callback functions that consume
            a negligible amount of time from the report

    Usage:
        Istantiate a profiler, and then register an event_handler with the
        syntax profiler(event_handler), which returns
        another object implementing the
        :class:`~pydgn.training.event.handler.EventHandler` interface
    """

    def __init__(self, threshold: float):
        # we filter out computation that takes a negligible amount of time from
        # the report (< threshold)
        self.threshold = threshold
        self.callback_elapsed = {}
        self.callback_calls = {}

    def __call__(self, event_handler: EventHandler) -> object:
        r"""
        Wraps a :class:`~pydgn.training.event.handler.EventHandler` object,
        so that whenever one of its callbacks is triggered the Profiler
        can compute and store statistics about the time required to execute it.

        Args:
            event_handler
                (:class:`~pydgn.training.event.handler.EventHandler`): the
                object implementing a subset of the callbacks defined in
                the :class:`~pydgn.training.event.handler.EventHandler`
                interface

        Returns:
            an object that transparently updates the profiler's state
            whenever a callback is triggered
        """
        self.callback_elapsed[event_handler.__class__.__name__] = {}
        self.callback_calls[event_handler.__class__.__name__] = {}

        def clock(func):
            """
            Decorator that takes a function and "clocks" it, by taking
            the time needed to execute it and updating
            statistics of the :obj:`Profiler` instance.
            """

            def clocked(*args, **kwargs):
                """
                Monitors the time to execute a callback function of some
                'obj':`EventHandler`.
                """
                class_name = event_handler.__class__.__name__
                callback_name = func.__name__
                t0 = time.time()
                # print(f'Callback {cls.__class__.__name__}: calling
                # {func.__name__} with args {args} and kwargs {kwargs}')
                result = func(*args, **kwargs)
                elapsed = time.time() - t0

                if callback_name in self.callback_elapsed[class_name]:
                    self.callback_elapsed[class_name][callback_name] += elapsed
                    self.callback_calls[class_name][callback_name] += 1
                else:
                    self.callback_elapsed[class_name][callback_name] = elapsed
                    self.callback_calls[class_name][callback_name] = 1

                return result

            return clocked

        class getattribute(type):
            """
            Taken from: (stackoverflow.com/questions/3155436/
            getattr-for-static-class-variables-in-python)
            We need to do this to invoke __getattr__ on a class
            (stackoverflow.com/questions/100003/what-are-metaclasses-in-python)
            TLDR We want the class ClockedCallback (which is an object itself)
            to have a method __getattr__. if we specified __getattr__ INSIDE
            ClockedCallback, rather than in its metaclass,
            the class ClockedCallback would not use the overridden __getattr__
            """

            def __getattr__(self, name):
                """
                See also class description. The method clocks the callback
                function whenever a specific attribute has to be retrieved.
                """
                return clock(
                    getattr(event_handler, name)
                )  # needed to implement callback calls with generic names

        # Relies on closures
        class ClockedCallback(metaclass=getattribute):
            """
            Dummy class to override method `__getattr__`
            """

            pass

        return ClockedCallback

    def report(self) -> str:
        r"""
        Builds a report string containing the statistics of the
        experiment accumulated so far.

        Returns:
            a string containing the report
        """
        total_time_experiment = 0.0
        profile_str = f'{"*" * 25} Profiler {"*" * 25} \n \n'
        profile_str += f"Threshold: {self.threshold} \n \n"

        for class_name, v in self.callback_elapsed.items():

            sorted_avg_elapsed = []

            for callback_name, v1 in v.items():
                n = self.callback_calls[class_name][callback_name]

                # Depending on when a KeyboardInterrupt is triggered
                if n == 0:
                    continue

                avg_elapsed = v1 / n

                if avg_elapsed > self.threshold:
                    bisect.insort(
                        sorted_avg_elapsed, (avg_elapsed, v1, callback_name)
                    )

            # Release resources
            v.clear()

            profile_str += f"{class_name} \n \n"

            for (avg_elapsed, total_elapsed, callback_name) in reversed(
                sorted_avg_elapsed
            ):
                total_time_experiment += total_elapsed
                seconds_elapsed = datetime.timedelta(seconds=total_elapsed)
                profile_str += (
                    f"\t {callback_name} --> "
                    f"Avg: {avg_elapsed} s, "
                    f"Total: {str(seconds_elapsed)} \n"
                )
            profile_str += "\n"

        total_time = datetime.timedelta(seconds=total_time_experiment)
        profile_str += f"Total time of the experiment: {str(total_time)} \n \n"
        profile_str += f'{"*" * 60}'
        return profile_str



import os
from pathlib import Path
from typing import Callable, List, Union, Tuple, Optional

import torch
from torch.utils.data import SequentialSampler
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader
from tqdm import tqdm

import pydgn
from pydgn.log.logger import Logger
from pydgn.model.interface import ModelInterface
from pydgn.static import *
from pydgn.training.callback.early_stopping import EarlyStopper
from pydgn.training.callback.engine_callback import EngineCallback
from pydgn.training.callback.gradient_clipping import GradientClipper
from pydgn.training.callback.metric import Metric
from pydgn.training.callback.optimizer import Optimizer
from pydgn.training.callback.plotter import Plotter
from pydgn.training.callback.scheduler import Scheduler
from pydgn.training.event.dispatcher import EventDispatcher
from pydgn.training.event.handler import EventHandler
from pydgn.training.event.state import State
from pydgn.training.profiler import Profiler


def log(msg, logger: Logger):
    """
    Logs a message using a logger
    """
    if logger is not None:
        logger.log(msg)


def reorder(obj: List[object], perm: List[int]):
    """
    Reorders a list of objects in ascending order according to the indices
    defined in permutation argument.

    Args:
        obj (List[object]): the list of objects
        perm (List[int]): the permutation

    Returns:
        The reordered list of objects
    """
    assert len(obj) == len(perm) and len(obj) > 0
    return [y for (x, y) in sorted(zip(perm, obj))]


class TrainingEngine(EventDispatcher):
    """
    This is the most important class when it comes to training a model.
    It implements the :class:`~pydgn.training.event.dispatcher.EventDispatcher`
    interface, which means that after registering some callbacks in a
    given order, it will proceed to trigger specific events that
    will result in the shared :class:`~pydgn.training.event.state.State`
    object being updated by the callbacks. Callbacks implement the EventHandler
    interface, and they receive the shared State object when any event is
    triggered. Knowing the order in which
    callbacks are called is important.
    The order is:

    * loss function
    * score function
    * gradient clipper
    * optimizer
    * early stopper
    * scheduler
    * plotter

    Args:
        engine_callback
            (Callable[...,
            :class:`~pydgn.training.callback.engine_callback.EngineCallback`]):
            the engine callback object to be used for data fetching and
            checkpoints (or even other purposes if necessary)
        model (:class:`~pydgn.model.interface.ModelInterface`):
            the model to be trained
        loss (:class:`~pydgn.training.callback.metric.Metric`):
            the loss to be used
        optimizer (:class:`~pydgn.training.callback.optimizer.Optimizer`):
            the optimizer to be used
        scorer (:class:`~pydgn.training.callback.metric.Metric`):
            the score to be used
        scheduler (:class:`~pydgn.training.callback.scheduler.Scheduler`):
            the scheduler to be used Default is ``None``.
        early_stopper
            (:class:`~pydgn.training.callback.early_stopping.EarlyStopper`):
                the early stopper to be used. Default is ``None``.
        gradient_clipper
            (:class:`~pydgn.training.callback.gradient_clipper.GradientClipper`
            ): the gradient clipper to be used. Default is ``None``.
        device (str): the device on which to train. Default is ``cpu``.
        plotter (:class:`~pydgn.training.callback.plotter.Plotter`):
            the plotter to be used. Default is ``None``.
        exp_path (str): the path of the experiment folder.
            Default is ``None`` but it is always instantiated.
        evaluate_every (int): the frequency of logging epoch results.
            Default is ``1``.
        eval_training (bool): whether to re-evaluate loss and scores on the
            training set after a training epoch. Defaults to False.
        store_last_checkpoint (bool): whether to store a checkpoint at
            the end of each epoch. Allows to resume training from last epoch.
            Default is ``False``.
        reset_eval_model_hidden_state (bool): [temporal graph learning]
            Used when we want to reset the state after performing previous
            inference. It should be ``False`` when we are dealing
            with a single temporal graph sequence, because we don't want to
            reset the hidden state after processing the previous
            [training/validation] time steps.
    """

    def __init__(
        self,
        engine_callback: Callable[..., EngineCallback],
        model: ModelInterface,
        loss: Metric,
        optimizer: Optimizer,
        scorer: Metric,
        scheduler: Scheduler = None,
        early_stopper: EarlyStopper = None,
        gradient_clipper: GradientClipper = None,
        device: str = "cpu",
        plotter: Plotter = None,
        exp_path: str = None,
        evaluate_every: int = 1,
        eval_training: bool = False,
        store_last_checkpoint: bool = False,
        reset_eval_model_hidden_state: bool = True,
    ):
        super().__init__()

        self.engine_callback = engine_callback
        self.model = model
        self.loss_fun = loss
        self.optimizer = optimizer
        self.score_fun = scorer
        self.scheduler = scheduler
        self.early_stopper = early_stopper
        self.gradient_clipper = gradient_clipper
        self.device = device
        self.plotter = plotter
        self.exp_path = exp_path
        self.evaluate_every = evaluate_every
        self.eval_training = eval_training
        self.store_last_checkpoint = store_last_checkpoint
        self.training = False
        self.logger = None

        # For dynamic graph learning
        self.reset_eval_model_hidden_state = reset_eval_model_hidden_state

        self.profiler = Profiler(threshold=1e-5)

        # Now register the callbacks (IN THIS ORDER, KNOWN TO THE USER)
        # Decorate with a profiler
        self.callbacks = [
            self.profiler(c)
            for c in [
                self.loss_fun,
                self.score_fun,
                self.gradient_clipper,
                self.optimizer,
                self.early_stopper,
                self.scheduler,
                self.plotter,
            ]
            if c is not None
        ]  # filter None callbacks

        # Add an Engine specific callback to profile different passages of
        # _loop
        self.callbacks.append(
            self.profiler(
                self.engine_callback(
                    store_last_checkpoint=self.store_last_checkpoint
                )
            )
        )

        for c in self.callbacks:
            self.register(c)

        self.state = State(
            self.model, self.optimizer, self.device
        )  # Initialize the state
        self.state.update(exp_path=self.exp_path)

    # TODO in general, there are no guarantees that y will be sufficient to
    # determine the task. This may have to be changed in the future.
    def _to_data_list(
        self, x: torch.Tensor, batch: torch.Tensor, y: Optional[torch.Tensor]
    ) -> List[Data]:
        """
        Converts model outputs back to a list of Data elements. Useful for
        incremental architectures.

        Args:
            x (:class:`torch.Tensor`): tensor holding information of different
                nodes/graphs embeddings
            batch (:class:`torch.Tensor`): the usual PyG batch tensor.
                Used to split node/graph embeddings graph-wise.
            y (:class:`torch.Tensor`): target labels, used to determine
                whether the task is graph prediction or node prediction.
                Can be ``None``.

        Returns:
            a list of PyG Data objects (with only ``x`` and ``y`` attributes)
        """
        data_list = []

        _, counts = torch.unique_consecutive(batch, return_counts=True)
        cumulative = torch.cumsum(counts, dim=0)

        if y is not None:
            is_graph_prediction = y.shape[0] == len(cumulative)
            y = y.unsqueeze(1) if y.dim() == 1 else y

            data_list.append(
                Data(
                    x=x[: cumulative[0]],
                    y=y[0].unsqueeze(0)
                    if is_graph_prediction
                    else y[: cumulative[0]],
                )
            )
            for i in range(1, len(cumulative)):
                g = Data(
                    x=x[cumulative[i - 1] : cumulative[i]],
                    y=y[i].unsqueeze(0)
                    if is_graph_prediction
                    else y[cumulative[i - 1] : cumulative[i]],
                )
                data_list.append(g)
        else:
            data_list.append(Data(x=x[: cumulative[0]]))
            for i in range(1, len(cumulative)):
                g = Data(x=x[cumulative[i - 1] : cumulative[i]])
                data_list.append(g)

        return data_list

    def _to_list(
        self,
        data_list: List[Data],
        embeddings: Union[Tuple[torch.Tensor], torch.Tensor],
        batch: torch.Tensor,
        edge_index: torch.Tensor,
        y: Optional[torch.Tensor],
    ) -> List[Data]:
        """
        Extends the ``data_list`` list of PyG Data objects with new samples.

        Args:
            data_list: a list of PyG Data objects (with only ``x`` and ``y``
                attributes)
            embeddings (:class:`torch.Tensor`): tensor holding information
                of different nodes/graphs embeddings
            batch (:class:`torch.Tensor`): the usual PyG batch tensor.
                Used to split node/graph embeddings graph-wise.
            edge_index:
            y (:class:`torch.Tensor`): target labels, used to determine
                whether the task is graph prediction or node prediction.
                Can be ``None``.

        Returns:
            a list of PyG Data objects (with only ``x`` and ``y`` attributes)
        """
        # Crucial: Detach the embeddings to free the computation graph!!
        if isinstance(embeddings, tuple):
            embeddings = tuple(
                [
                    e.detach().cpu() if e is not None else None
                    for e in embeddings
                ]
            )
        elif isinstance(embeddings, torch.Tensor):
            embeddings = embeddings.detach().cpu()
        else:
            raise NotImplementedError(
                "Embeddings not understood, should be torch.Tensor or "
                "Tuple of torch.Tensor"
            )

        # Convert embeddings back to a list of torch_geometric Data objects
        # Needs also y information to (possibly) use them as a tensor dataset
        # CRUCIAL: remember, the loader could have shuffled the data, so we
        # need to carry y around as well
        if data_list is None:
            data_list = []
        data_list.extend(self._to_data_list(embeddings, batch, y))
        return data_list

    def _num_targets(self, targets: torch.Tensor) -> int:
        """
        Computes the number of targets (**different from the dimension of
        each target**).

        Args:
            targets (:class:`torch.Tensor`: the ground truth tensor

        Returns:
             an integer with the number of targets to predict
        """
        if targets is None:
            return 0
        assert isinstance(
            targets, torch.Tensor
        ), "Expecting a tensor as target"
        num_targets = targets.shape[0]
        return num_targets

    def set_device(self):
        """
        Moves the model and the loss metric to the proper device.
        """
        self.model.to(self.device)
        self.loss_fun.to(self.device)

    def set_training_mode(self):
        """
        Sets the model and the internal state in ``TRAINING`` mode
        """
        self.model.train()
        self.training = True

    def set_eval_mode(self):
        """
        Sets the model and the internal state in ``EVALUATION`` mode
        """
        self.model.eval()
        self.training = False

    def _loop_helper(self):
        """
        Helper function that loops over the data.
        """
        # Move data to device
        data = self.state.batch_input
        if isinstance(data, list):
            # used by incremental construction
            data = [d.to(self.device) for d in data]
            _data = data[0]
        else:
            # standard case
            data = data.to(self.device)
            _data = data
        batch_idx, edge_index, targets = _data.batch, _data.edge_index, _data.y

        # Helpful when you need to access again the input batch, e.g for some
        # continual learning strategy
        self.state.update(batch_input=data)
        self.state.update(batch_targets=targets)

        num_graphs = _data.num_graphs
        num_nodes = _data.num_nodes
        self.state.update(batch_num_graphs=num_graphs)
        self.state.update(batch_num_nodes=num_nodes)

        num_targets = self._num_targets(targets)
        self.state.update(batch_num_targets=num_targets)

        # Reset batch_loss and batch_score states (ow accumulation of batch
        # results breaks for temporal graph learning)
        self.state.update(batch_loss=None)
        self.state.update(batch_score=None)

        if self.training:
            self._dispatch(EventHandler.ON_TRAINING_BATCH_START, self.state)
        else:
            self._dispatch(EventHandler.ON_EVAL_BATCH_START, self.state)

        self._dispatch(EventHandler.ON_FORWARD, self.state)

        # EngineCallback will store the outputs in state.batch_outputs
        output = self.state.batch_outputs

        # Change into tuple if not, loss and score expect a tuple
        # where the first element consists of the model's predictions
        if not isinstance(output, tuple):
            output = (output,)
        else:
            if len(output) > 1 and self.state.return_node_embeddings:
                # Embeddings should be in position 2 of the output
                embeddings = output[1]

                data_list = self._to_list(
                    self.state.epoch_data_list,
                    embeddings,
                    batch_idx,
                    edge_index,
                    targets,
                )

                # I am extending the data list, not replacing! Hence the name
                # "epoch" data list
                self.state.update(epoch_data_list=data_list)

        self._dispatch(EventHandler.ON_COMPUTE_METRICS, self.state)

        if self.training:
            self._dispatch(EventHandler.ON_BACKWARD, self.state)
            self._dispatch(EventHandler.ON_TRAINING_BATCH_END, self.state)
        else:
            self._dispatch(EventHandler.ON_EVAL_BATCH_END, self.state)

    # loop over all data (i.e. computes an epoch)
    def _loop(self, loader: DataLoader):
        """
        Main method that computes a pass over the dataset using the data
        loader provided.

        Args:
            loader (:class:`torch_geometric.loader.DataLoader`):
                the loader to be used
        """
        # Reset epoch state (DO NOT REMOVE)
        self.state.update(epoch_data_list=None)
        self.state.update(epoch_loss=None)
        self.state.update(epoch_score=None)
        self.state.update(loader_iterable=iter(loader))

        # Loop over data
        for id_batch in tqdm(range(len(loader)),
                             desc=f'Epoch {self.state.epoch+1}, {self.state.set} set',
                             unit='batch',
                             disable=not self.logger.debug,
                             leave=False):
            self.state.update(id_batch=id_batch)
            # EngineCallback will store fetched data in state.batch_input
            self._dispatch(EventHandler.ON_FETCH_DATA, self.state)

            self._loop_helper()

    def _train(self, loader):
        """
        Implements a loop over the data in training mode
        """
        self.set_training_mode()

        self._dispatch(EventHandler.ON_TRAINING_EPOCH_START, self.state)

        self._loop(loader)

        self._dispatch(EventHandler.ON_TRAINING_EPOCH_END, self.state)

        assert self.state.epoch_loss is not None
        loss, score = self.state.epoch_loss, self.state.epoch_score
        return loss, score, None

    def infer(
        self, loader: DataLoader, set: str
    ) -> Tuple[dict, dict, List[Data]]:
        """
        Performs an evaluation step on the data.

        Args:
            loader (:class:`torch_geometric.loader.DataLoader`):
                the loader to be used
            set (str): the type of dataset being used, can be ``TRAINING``,
                ``VALIDATION`` or ``TEST`` (as defined in ``pydgn.static``)

        Returns:
             a tuple (loss dict, score dict, list of
             :class:`torch_geometric.data.Data` objects with ``x`` and
             ``y`` attributes only).
             The data list can be used, for instance, in
             semi-supervised experiments or in incremental architectures
        """
        self.set_eval_mode()
        self.state.update(set=set)

        self._dispatch(EventHandler.ON_EVAL_EPOCH_START, self.state)

        with torch.no_grad():
            self._loop(loader)  # state has been updated

        self._dispatch(EventHandler.ON_EVAL_EPOCH_END, self.state)

        assert self.state.epoch_loss is not None
        loss, score, data_list = (
            self.state.epoch_loss,
            self.state.epoch_score,
            self.state.epoch_data_list,
        )

        # Add the main loss we want to return as a special key
        main_loss_name = self.loss_fun.get_main_metric_name()
        loss[MAIN_LOSS] = loss[main_loss_name]

        # Add the main score we want to return as a special key
        # Needed by the experimental evaluation framework
        main_score_name = self.score_fun.get_main_metric_name()
        score[MAIN_SCORE] = score[main_score_name]

        # If samples been shuffled by the data loader, i.e., in the
        # last inference phase of the engine, use PyDGN RandomSampler to
        # recover the permutation and store the embedding data list as in
        # the original split. This way we recover the same order of the
        # [train/val/test] idx list in the split file. This is useful for
        # subsequent experiments that use the data list and must be
        # consistent, and it is better in general to avoid confusion.
        if data_list is not None and loader.sampler is not None:
            # if SequentialSampler then shuffle was false
            if not isinstance(loader.sampler, SequentialSampler):
                assert isinstance(
                    loader.sampler, pydgn.data.sampler.RandomSampler
                ), (
                    "Training Engine requires a pydgn.data.sampler."
                    "RandomSampler as the sampler of the loader when "
                    "shuffle=True. Please see the documentation of "
                    "DataProvider and the _get_loader method."
                )
                data_list = reorder(data_list, loader.sampler.permutation)

        return loss, score, data_list

    def train(
        self,
        train_loader: DataLoader,
        validation_loader: DataLoader = None,
        test_loader: DataLoader = None,
        max_epochs: int = 100,
        zero_epoch: bool = False,
        logger: Logger = None,
    ) -> Tuple[
        dict, dict, List[Data], dict, dict, List[Data], dict, dict, List[Data]
    ]:
        """
        Trains the model and regularly evaluates on validation and test data
        (if given). May perform early stopping and checkpointing.

        Args:
            train_loader (:class:`torch_geometric.loader.DataLoader`):
                the DataLoader associated with training data
            validation_loader (:class:`torch_geometric.loader.DataLoader`):
                the DataLoader associated with validation data, if any
            test_loader (:class:`torch_geometric.loader.DataLoader`):
                the DataLoader associated with test data, if any
            max_epochs (int): maximum number of training epochs.
                Default is ``100``
            zero_epoch: if ``True``, starts again from epoch 0 and resets
                optimizer and scheduler states. Default is ``False``
            logger: the logger

        Returns:
             a tuple (train_loss, train_score, train_embeddings,
             validation_loss, validation_score, validation_embeddings,
             test_loss, test_score, test_embeddings)
        """
        self.logger = logger

        try:
            # Initialize variables
            val_loss, val_score, val_embeddings_tuple = None, None, None
            test_loss, test_score, test_embeddings_tuple = None, None, None

            self.set_device()

            # Restore training from last checkpoint if possible!
            ckpt_filename = Path(self.exp_path, LAST_CHECKPOINT_FILENAME)
            best_ckpt_filename = Path(self.exp_path, BEST_CHECKPOINT_FILENAME)
            is_early_stopper_ckpt = (
                self.early_stopper.checkpoint
                if self.early_stopper is not None
                else False
            )
            # If one changes the options in the config file, the existence of
            # a checkpoint is not enough to
            # decide whether to resume training or not!
            if os.path.exists(ckpt_filename) and (
                self.store_last_checkpoint or is_early_stopper_ckpt
            ):
                self._restore_checkpoint_and_best_results(
                    ckpt_filename, best_ckpt_filename, zero_epoch
                )
                log(
                    f"START AGAIN FROM EPOCH {self.state.initial_epoch}",
                    logger,
                )

            self._dispatch(EventHandler.ON_FIT_START, self.state)

            # In case we already have a trained model
            epoch = self.state.initial_epoch

            # Loop over the entire dataset dataset
            for epoch in range(self.state.initial_epoch, max_epochs):
                self.state.update(epoch=epoch)
                self.state.update(return_node_embeddings=False)

                # [temporal] Initialize the hidden state of the model
                # for a training epoch
                self.state.update(last_hidden_state=None)

                self._dispatch(EventHandler.ON_EPOCH_START, self.state)

                self.state.update(set=TRAINING)
                train_loss, train_score, _ = self._train(train_loader)

                # Update state with epoch results
                epoch_results = {LOSSES: {}, SCORES: {}}

                if self.eval_training:
                    # needs to reset the temporal state in any case since
                    # we restart the inference from the training set
                    # otherwise we keep the temporal hidden state as it is

                    # [temporal] Initialize the hidden state of the model
                    # again before inference
                    self.state.update(last_hidden_state=None)

                if (
                    (epoch + 1) >= self.evaluate_every
                    and (epoch + 1) % self.evaluate_every == 0
                ) or epoch == 0:

                    if self.eval_training:
                        # Compute training output (necessary because on_backward
                        # has been called)
                        train_loss, train_score, _ = self.infer(
                            train_loader, TRAINING
                        )

                    if self.reset_eval_model_hidden_state:
                        self.state.update(last_hidden_state=None)

                    # Compute validation output
                    if validation_loader is not None:
                        val_loss, val_score, _ = self.infer(
                            validation_loader, VALIDATION
                        )

                    if self.reset_eval_model_hidden_state:
                        self.state.update(last_hidden_state=None)

                    # Compute test output for visualization purposes only (e.g.
                    # to debug an incorrect data split for link prediction)
                    if test_loader is not None:
                        test_loss, test_score, _ = self.infer(
                            test_loader, TEST
                        )

                    epoch_results[LOSSES].update(
                        {f"{TRAINING}_{k}": v for k, v in train_loss.items()}
                    )
                    epoch_results[SCORES].update(
                        {f"{TRAINING}_{k}": v for k, v in train_score.items()}
                    )

                    if validation_loader is not None:
                        epoch_results[LOSSES].update(
                            {
                                f"{VALIDATION}_{k}": v
                                for k, v in val_loss.items()
                            }
                        )
                        epoch_results[SCORES].update(
                            {
                                f"{VALIDATION}_{k}": v
                                for k, v in val_score.items()
                            }
                        )
                        val_msg_str = (
                            f", VL loss: {val_loss} VL score: {val_score}"
                        )
                    else:
                        val_msg_str = ""

                    if test_loader is not None:
                        epoch_results[LOSSES].update(
                            {f"{TEST}_{k}": v for k, v in test_loss.items()}
                        )
                        epoch_results[SCORES].update(
                            {f"{TEST}_{k}": v for k, v in test_score.items()}
                        )
                        test_msg_str = (
                            f", TE loss: {test_loss} TE score: {test_score}"
                        )
                    else:
                        test_msg_str = ""

                    # Log performances
                    msg = (
                        f"Epoch: {epoch + 1}, TR loss: {train_loss} "
                        f"TR score: {train_score}" + val_msg_str + test_msg_str
                    )
                    log(msg, logger)

                # Update state with the result of this epoch
                self.state.update(epoch_results=epoch_results)

                # We can apply early stopping here
                self._dispatch(EventHandler.ON_EPOCH_END, self.state)

                if self.state.stop_training:
                    log(f"Stopping at epoch {self.state.epoch}.", logger)
                    break

            # Needed to indicate that training has ended
            self.state.update(stop_training=True)

            # We reached the max # of epochs, get best scores from the early
            # stopper (if any) and restore best model
            if self.early_stopper is not None:
                ber = self.state.best_epoch_results
                # Restore the model according to the best validation score!
                self.model.load_state_dict(ber[MODEL_STATE])
            else:
                self.state.update(best_epoch_results={BEST_EPOCH: epoch})
                ber = self.state.best_epoch_results

            self.state.update(return_node_embeddings=True)

            # Initialize the state of the model again before the final
            # evaluation
            self.state.update(last_hidden_state=None)

            # Compute training output
            train_loss, train_score, train_embeddings_tuple = self.infer(
                train_loader, TRAINING
            )
            # ber[f'{TRAINING}_loss'] = train_loss
            ber[f"{TRAINING}{EMB_TUPLE_SUBSTR}"] = train_embeddings_tuple
            ber.update({f"{TRAINING}_{k}": v for k, v in train_loss.items()})
            ber.update({f"{TRAINING}_{k}": v for k, v in train_score.items()})

            if self.reset_eval_model_hidden_state:
                self.state.update(last_hidden_state=None)

            # Compute validation output
            if validation_loader is not None:
                val_loss, val_score, val_embeddings_tuple = self.infer(
                    validation_loader, VALIDATION
                )
                # ber[f'{VALIDATION}_loss'] = val_loss
                ber[f"{VALIDATION}{EMB_TUPLE_SUBSTR}"] = val_embeddings_tuple
                ber.update({f"{TRAINING}_{k}": v for k, v in val_loss.items()})
                ber.update(
                    {f"{TRAINING}_{k}": v for k, v in val_score.items()}
                )

            if self.reset_eval_model_hidden_state:
                self.state.update(last_hidden_state=None)

            # Compute test output
            if test_loader is not None:
                test_loss, test_score, test_embeddings_tuple = self.infer(
                    test_loader, TEST
                )
                # ber[f'{TEST}_loss'] = test_loss
                ber[f"{TEST}{EMB_TUPLE_SUBSTR}"] = test_embeddings_tuple
                ber.update({f"{TEST}_{k}": v for k, v in test_loss.items()})
                ber.update({f"{TEST}_{k}": v for k, v in test_score.items()})

            self._dispatch(EventHandler.ON_FIT_END, self.state)

            self.state.update(return_node_embeddings=False)

            log(
                f"Chosen is Epoch {ber[BEST_EPOCH] + 1} "
                f"TR loss: {train_loss} TR score: {train_score}, "
                f"VL loss: {val_loss} VL score: {val_score} "
                f"TE loss: {test_loss} TE score: {test_score}",
                logger,
            )

            self.state.update(set=None)

        except (KeyboardInterrupt, RuntimeError, FileNotFoundError) as e:
            report = self.profiler.report()
            print(str(e))
            log(str(e), logger)
            log(report, logger)
            raise e
            exit(0)

        # Log profile results
        report = self.profiler.report()
        log(report, logger)

        return (
            train_loss,
            train_score,
            train_embeddings_tuple,
            val_loss,
            val_score,
            val_embeddings_tuple,
            test_loss,
            test_score,
            test_embeddings_tuple,
        )

    def _restore_checkpoint_and_best_results(
        self, ckpt_filename, best_ckpt_filename, zero_epoch
    ):
        """
        Restores the (best or last) checkpoint from a given file, and loads
        the best results so far into the state if any.
        """
        # When changing exp config from cuda to cpu, cuda will not be available
        # to pytorch (due to Ray management of resources). Hence, we need to
        # specify explicitly the map location as cpu. The other way around
        # (cpu to cuda) poses no problem since GPUs are visible.
        ckpt_dict = torch.load(
            ckpt_filename, map_location="cpu" if self.device == "cpu" else None
        )

        self.state.update(
            initial_epoch=int(ckpt_dict[EPOCH]) + 1 if not zero_epoch else 0
        )
        self.state.update(stop_training=ckpt_dict[STOP_TRAINING])

        model_state = ckpt_dict[MODEL_STATE]

        # Needed only when moving from cpu to cuda (due to changes in config
        # file). Move all parameters to cuda.
        for param in model_state.keys():
            model_state[param] = model_state[param].to(self.device)

        self.model.load_state_dict(model_state)

        if not zero_epoch:
            optimizer_state = ckpt_dict[OPTIMIZER_STATE]
            self.state.update(optimizer_state=optimizer_state)

        if os.path.exists(best_ckpt_filename):
            best_ckpt_dict = torch.load(
                best_ckpt_filename,
                map_location="cpu" if self.device == "cpu" else None,
            )
            self.state.update(best_epoch_results=best_ckpt_dict)

        if self.scheduler is not None and not zero_epoch:
            scheduler_state = ckpt_dict[SCHEDULER_STATE]
            assert scheduler_state is not None
            self.state.update(scheduler_state=scheduler_state)


class DataStreamTrainingEngine(TrainingEngine):
    """
    Class that handles a stream of graphs that could end at any moment.
    """

    # loop over all data (i.e. computes an epoch)
    def _loop(self, loader: DataLoader):
        """
        Compared to superclass version, handles the issue of a stream of data
        that could end at any moment. This is done using an additional
        boolean state variable.
        """
        # Reset epoch state (DO NOT REMOVE)
        self.state.update(epoch_data_list=None)
        self.state.update(epoch_loss=None)
        self.state.update(epoch_score=None)
        self.state.update(loader_iterable=iter(loader))

        # Loop over data
        id_batch = 0
        self.state.update(stop_fetching=False)
        while not self.state.stop_fetching:

            self.state.update(id_batch=id_batch)
            id_batch += 1

            # EngineCallback will store fetched data in state.batch_input
            self._dispatch(EventHandler.ON_FETCH_DATA, self.state)

            if self.state.stop_fetching:
                return

            self._loop_helper()


class GraphSequenceTrainingEngine(TrainingEngine):
    """
    Assumes that the model can return None predictions (i.e., output[0] below)
    whenever no values should be predicted according to the mask field in the
    Data object representing a snapshot
    """

    def _loop_helper(self):
        """
        Compared to superclass version, aas the batch is composed of a list
        of snapshots, where each snapshot is a graph object, we add an inner
        loop to pass one snapshot at a time.
        """
        # data is a list of snapshots, one per time step
        data = self.state.batch_input

        # Reset batch_loss and batch_score states (ow accumulation of batch
        # results breaks for temporal graph learning)
        self.state.update(batch_loss=None)
        self.state.update(batch_score=None)

        if self.training:
            self._dispatch(EventHandler.ON_TRAINING_BATCH_START, self.state)
        else:
            self._dispatch(EventHandler.ON_EVAL_BATCH_START, self.state)

        for t, snapshot in enumerate(data):
            self.state.update(time_step=t)

            if isinstance(snapshot, list):
                # used by incremental construction
                snapshot = [d.to(self.device) for d in snapshot]
                _snapshot = snapshot[0]
            else:
                # standard case
                snapshot = snapshot.to(self.device)
                _snapshot = snapshot

            batch_idx, edge_index, targets = (
                _snapshot.batch,
                _snapshot.edge_index,
                _snapshot.y,
            )

            # Helpful when you need to access again the input batch, e.g for
            # some continual learning strategy
            self.state.update(batch_input=_snapshot)
            self.state.update(batch_targets=targets)

            num_graphs = _snapshot.num_graphs
            num_nodes = _snapshot.num_nodes
            self.state.update(batch_num_graphs=num_graphs)
            self.state.update(batch_num_nodes=num_nodes)

            num_targets = self._num_targets(targets)
            self.state.update(batch_num_targets=num_targets)

            self._dispatch(EventHandler.ON_FORWARD, self.state)

            # EngineCallback will store the outputs in state.batch_outputs
            output = self.state.batch_outputs

            # Update previous hidden states
            self.state.update(last_hidden_state=output[1])

            # Change into tuple if not, loss and score expect a tuple
            # where the first element consists of the model's predictions
            if not isinstance(output, tuple):
                output = (output,)
            else:
                if len(output) > 1 and self.state.return_node_embeddings:
                    # Embeddings should be in position 2 of the output
                    embeddings = output[1]

                    data_list = self._to_list(
                        self.state.epoch_data_list,
                        embeddings,
                        batch_idx,
                        edge_index,
                        targets,
                    )

                    # I am extending the data list, not replacing! Hence the
                    # name "epoch" data list
                    self.state.update(epoch_data_list=data_list)

            self._dispatch(EventHandler.ON_COMPUTE_METRICS, self.state)

        if self.training:
            self._dispatch(EventHandler.ON_BACKWARD, self.state)

            # IMPORTANT: after backward detach the last hidden state
            last_hidden_state = self.state.last_hidden_state.detach()
            self.state.update(last_hidden_state=last_hidden_state)

            self._dispatch(EventHandler.ON_TRAINING_BATCH_END, self.state)
        else:
            self._dispatch(EventHandler.ON_EVAL_BATCH_END, self.state)


class LinkPredictionSingleGraphEngine(TrainingEngine):
    """
    Specific engine for link prediction tasks. Here, we expect target values
    in the form of tuples: ``(_, pos_edges, neg_edges)``, where ``pos_edges``
    and ``neg_edges`` have been generated by the splitter and provided
    by the data provider.
    """

    def _to_data_list(self, x, batch, y):
        """
        Converts model outputs back to a list of Data elements.
        Useful for incremental architectures.

        Args:
            x (:class:`torch.Tensor`): tensor holding information of
                different nodes/graphs embeddings
            batch (:class:`torch.Tensor`): the usual PyG batch tensor.
                Used to split node/graph embeddings graph-wise.
            y (:class:`torch.Tensor`): target labels, used to determine
                whether the task is graph prediction or node prediction.
                Can be ``None``.

        Returns:
            a list of PyG Data objects (with only ``x`` and ``y`` attributes)
        """
        data_list = []
        # Return node embeddings and their original class, if any (or dumb
        # value which is required nonetheless)
        y = y[0]
        data_list.append(Data(x=x, y=y[0]))
        return data_list

    def _to_list(self, data_list, embeddings, batch, edge_index, y):
        """
        Extends the ``data_list`` list of PyG Data objects with new samples.

        Args:
            data_list: a list of PyG Data objects (with only ``x``
                and ``y`` attributes)
            embeddings (:class:`torch.Tensor`): tensor holding information
                of different nodes/graphs embeddings
            batch (:class:`torch.Tensor`): the usual PyG batch tensor.
                Used to split node/graph embeddings graph-wise.
            edge_index:
            y (:class:`torch.Tensor`): target labels, used to determine
                whether the task is graph prediction or node prediction.
                Can be ``None``.

        Returns:
            a list of PyG Data objects (with only ``x`` and ``y`` attributes)
        """
        assert isinstance(
            y, list
        ), "Expecting a list of (_, pos_edges, neg_edges)"

        # Crucial: Detach the embeddings to free the computation graph!!
        if isinstance(embeddings, tuple):
            embeddings = tuple(
                [
                    e.detach().cpu() if e is not None else None
                    for e in embeddings
                ]
            )
        elif isinstance(embeddings, torch.Tensor):
            embeddings = embeddings.detach().cpu()
        else:
            raise NotImplementedError(
                "Embeddings not understood, should be torch.Tensor or "
                "Tuple of torch.Tensor"
            )

        # Convert embeddings back to a list of torch_geometric Data objects
        # Needs also y information to (possibly) use them as a tensor dataset
        # (remember, the loader could have shuffled the data)
        if data_list is None:
            data_list = []
        data_list.extend(self._to_data_list(embeddings, batch, y))
        return data_list

    def _num_targets(self, targets):
        """
        Computes the number of targets as the positive links + negative links
        """
        assert isinstance(
            targets, list
        ), "Expecting a list of (_, pos_edges, neg_edges)"
        # positive links + negative links provided separately
        num_targets = targets[1].shape[1] + targets[2].shape[1]
        # Just a single graph for link prediction
        return num_targets



from typing import List, Optional, Tuple

import torch
from torch import nn
from torch_geometric.data import Batch
from torch_geometric.nn import global_add_pool

from pydgn.model.interface import ModelInterface


class ToyMLP(ModelInterface):
    """
    A toy MLP model used to test the library. Technically,
    a DGN that ignores the adjacency matrix.
    """

    def __init__(
        self,
        dim_node_features,
        dim_edge_features,
        dim_target,
        readout_class,
        config,
    ):
        super().__init__(
            dim_node_features,
            dim_edge_features,
            dim_target,
            readout_class,
            config,
        )

        dim_embedding = config["dim_embedding"]
        self.W = nn.Linear(dim_node_features, dim_target, bias=True)

    def forward(
        self, data: Batch
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        """
        Implements an MLP (structure agnostic baseline)

        Args:
            data (torch_geometric.data.Batch): a batch of graphs

        Returns:
            a tuple (output, node_embedddings)
        """
        x, batch = data.x, data.batch

        hg = global_add_pool(x, batch)
        return self.W(hg), x



import math
import random
from typing import Union, Callable, List

import numpy as np
import torch
import torch_geometric.loader
from torch.utils.data import Subset
from torch_geometric.data import Data, Batch
from torch_geometric.loader.dataloader import Collater

import pydgn.data.dataset
from pydgn.data.dataset import DatasetInterface
from pydgn.data.dataset import TemporalDatasetInterface
from pydgn.data.sampler import RandomSampler
from pydgn.data.splitter import (
    Splitter,
    LinkPredictionSingleGraphSplitter,
    SingleGraphSplitter,
)
from pydgn.data.util import load_dataset


def seed_worker(exp_seed, worker_id):
    r"""
    Used to set a different, but reproducible, seed for all data-retriever
    workers. Without this, all workers will retrieve the data in the same
    order (important for Iterable-style datasets).

    Args:
        exp_seed (int): base seed to be used for reproducibility
        worker_id (int): id number of the worker
    """
    np.random.seed(exp_seed + worker_id)
    random.seed(exp_seed + worker_id)
    torch.manual_seed(exp_seed + worker_id)
    torch.cuda.manual_seed(exp_seed + worker_id)


class DataProvider:
    r"""
    A DataProvider object retrieves the correct data according to the external
    and internal data splits. It can be additionally used to augment the data,
    or to create a specific type of data loader. The base class does nothing
    special, but here is where the i-th element of a dataset could be
    pre-processed before constructing the mini-batches.

    IMPORTANT: if the dataset is to be shuffled, you MUST use a
    :class:`pydgn.data.sampler.RandomSampler` object to determine the
    permutation.

    Args:
        data_root (str): the path of the root folder in which data is stored
        splits_filepath (str): the filepath of the splits. with additional
            metadata
        dataset_class
            (Callable[...,:class:`pydgn.data.dataset.DatasetInterface`]):
            the class of the dataset
        data_loader_class
            (Union[Callable[...,:class:`torch.utils.data.DataLoader`],
            Callable[...,:class:`torch_geometric.loader.DataLoader`]]):
            the class of the data loader to use
        data_loader_args (dict): the arguments of the data loader
        dataset_name (str): the name of the dataset
        outer_folds (int): the number of outer folds for risk assessment.
            1 means hold-out, >1 means k-fold
        inner_folds (int): the number of outer folds for model selection.
            1 means hold-out, >1 means k-fold

    """

    def __init__(
        self,
        data_root: str,
        splits_filepath: str,
        dataset_class: Callable[..., pydgn.data.dataset.DatasetInterface],
        dataset_name: str,
        data_loader_class: Union[
            Callable[..., torch.utils.data.DataLoader],
            Callable[..., torch_geometric.loader.DataLoader],
        ],
        data_loader_args: dict,
        outer_folds: int,
        inner_folds: int,
    ):

        self.exp_seed = None

        self.data_root = data_root
        self.dataset_class = dataset_class
        self.dataset_name = dataset_name

        self.data_loader_class = data_loader_class
        self.data_loader_args = data_loader_args

        self.outer_folds = outer_folds
        self.inner_folds = inner_folds

        self.outer_k = None
        self.inner_k = None

        self.splits_filepath = splits_filepath
        self.splitter = None

        # use this to avoid instantiating multiple versions of the same
        # dataset when no run-time
        # specific arguments are needed
        self.dataset = None

        self.dim_node_features = None
        self.dim_edge_features = None
        self.dim_target = None

    def set_exp_seed(self, seed: int):
        r"""
        Sets the experiment seed to give to the DataLoader.
        Helps with reproducibility.

        Args:
            seed (int): id of the seed
        """
        self.exp_seed = seed

    def set_outer_k(self, k: int):
        r"""
        Sets the parameter k of the `risk assessment` procedure.
        Called by the evaluation modules to load the correct data subset.

        Args:
            k (int): the id of the fold, ranging from 0 to K-1.
        """
        self.outer_k = k

    def set_inner_k(self, k):
        r"""
        Sets the parameter k of the `model selection` procedure. Called by the
        evaluation modules to load the correct subset of the data.

        Args:
            k (int): the id of the fold, ranging from 0 to K-1.
        """
        self.inner_k = k

    def _get_splitter(self) -> Splitter:
        """
        Instantiates the splitter with the parameters stored
        in the file ``self.splits_filepath``

        Returns:
            a :class:`~pydgn.data.splitter.Splitter` object
        """
        if self.splitter is None:
            self.splitter = Splitter.load(self.splits_filepath)
        return self.splitter

    def _get_dataset(self, **kwargs: dict) -> DatasetInterface:
        """
        Instantiates the dataset. Relies on the parameters stored in
        the ``dataset_kwargs.pt`` file.

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset. Not used in the base version

        Returns:
            a :class:`~pydgn.data.dataset.DatasetInterface` object
        """
        assert "shuffle" not in kwargs, (
            "Your implementation of _get_loader should remove `shuffle` "
            "from kwargs before calling _get_dataset"
        )
        assert "batch_size" not in kwargs, (
            "Your implementation of _get_loader should remove `batch_size` "
            "from kwargs before calling _get_dataset"
        )

        if kwargs is not None and len(kwargs) != 0:
            # we probably need to pass run-time specific parameters,
            # so load the dataset in memory again
            # an example is the subset of urls in Iterable style datasets
            dataset = load_dataset(
                self.data_root, self.dataset_name, self.dataset_class, **kwargs
            )
        else:
            if self.dataset is None:
                dataset = load_dataset(
                    self.data_root, self.dataset_name, self.dataset_class
                )
                self.dataset = dataset
            else:
                dataset = self.dataset

        self.dim_node_features = dataset.dim_node_features
        self.dim_edge_features = dataset.dim_edge_features
        self.dim_target = dataset.dim_target

        return dataset

    def _get_loader(
        self, indices: list, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Instantiates the data loader.

        Args:
            indices (sequence): Indices in the whole set selected for subset
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object
        """
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        dataset: DatasetInterface = self._get_dataset(**kwargs)
        dataset = Subset(dataset, indices)

        assert (
            self.exp_seed is not None
        ), "DataLoader's seed has not been specified! Is this a bug?"

        # no need to set worker seed in map-stye dataset, see pytorch doc about reproducibility
        # this would also cause transforms based on random sampling to behave in the same way every time
        # kwargs["worker_init_fn"] = lambda worker_id: seed_worker(
        #     worker_id, self.exp_seed
        # )

        kwargs.update(self.data_loader_args)

        if shuffle is True:
            sampler = RandomSampler(dataset)
            dataloader = self.data_loader_class(
                dataset, sampler=sampler, batch_size=batch_size, **kwargs
            )
        else:
            dataloader = self.data_loader_class(
                dataset, shuffle=False, batch_size=batch_size, **kwargs
            )

        return dataloader

    def get_inner_train(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for model selection associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        indices = splitter.inner_folds[self.outer_k][self.inner_k].train_idxs
        return self._get_loader(indices, **kwargs)

    def get_inner_val(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for model selection associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        indices = splitter.inner_folds[self.outer_k][self.inner_k].val_idxs
        return self._get_loader(indices, **kwargs)

    def get_outer_train(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()

        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        return self._get_loader(train_indices, **kwargs)

    def get_outer_val(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for risk assessment associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        val_indices = splitter.outer_folds[self.outer_k].val_idxs
        return self._get_loader(val_indices, **kwargs)

    def get_outer_test(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the test set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        indices = splitter.outer_folds[self.outer_k].test_idxs
        return self._get_loader(indices, **kwargs)

    def get_dim_node_features(self) -> int:
        r"""
        Returns the number of node features of the dataset

        Returns:
            the value of the property ``dim_node_features`` in the dataset

        """
        if self.dim_node_features is None:
            raise Exception(
                "You should first initialize the dataset "
                "by creating a data loader!"
            )
        return self.dim_node_features

    def get_dim_edge_features(self) -> int:
        r"""
        Returns the number of node features of the dataset

        Returns:
            the value of the property ``dim_edge_features`` in the dataset

        """
        if self.dim_edge_features is None:
            raise Exception(
                "You should first initialize the dataset "
                "by creating a data loader!"
            )
        return self.dim_edge_features

    def get_dim_target(self) -> int:
        r"""
        Returns the dimension of the target for the task

        Returns:
            the value of the property ``dim_target`` in the dataset

        """
        if self.dim_target is None:
            raise Exception(
                "You should first initialize the dataset "
                "by creating a data loader!"
            )
        return self.dim_target


class IterableDataProvider(DataProvider):
    r"""
    A DataProvider object that allows to fetch data from an Iterable-style
    Dataset (see :class:`pydgn.data.dataset.IterableDatasetInterface`).
    """

    def _get_loader(
        self, indices: list, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Instantiates the data loader, passing to the dataset an additional
        `url_indices` argument with the indices to fetch. This is because
        each time this method is called with different indices a separate
        instance of the dataset is called.

        Args:
            indices (sequence): Indices in the whole set selected for subset
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object
        """
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        # we will overwrite the dataset each time the loader is called
        dataset = self._get_dataset(**{"url_indices": indices})

        if shuffle:
            dataset.shuffle_urls(True)
            dataset.shuffle_urls_elements(True)

        assert (
            self.exp_seed is not None
        ), "DataLoader's seed has not been specified! Is this a bug?"

        # Define a worker_init_fn that configures each dataset copy differently
        # this is called only when num_workers is set to a value > 0
        def worker_init_fn(worker_id: int):
            """
            Set the seeds for the worker and computes the range of
            samples ids to fetch.
            """
            worker_info = torch.utils.data.get_worker_info()
            num_workers = worker_info.num_workers
            assert num_workers > 0

            # Set the random seed
            seed_worker(worker_id, self.exp_seed)

            # Get the dataset and overall length
            dataset = (
                worker_info.dataset
            )  # the dataset copy in this worker process
            dataset_length = len(
                dataset
            )  # dynamic, already refers to the subset of urls!

            per_worker = int(
                math.ceil((dataset_length) / float(worker_info.num_workers))
            )

            start = worker_id * per_worker
            end = worker_id * per_worker + per_worker

            # configure the dataset to only process the split workload
            dataset.splice(start, end)

        kwargs.update(self.data_loader_args)

        dataloader = self.data_loader_class(
            dataset,
            sampler=None,
            collate_fn=Collater(None, None),
            worker_init_fn=worker_init_fn,
            batch_size=batch_size,
            **kwargs,
        )
        return dataloader


class SingleGraphSequenceDataProvider(DataProvider):
    """
    This class is responsible for building the dynamic dataset at runtime.
    """

    @classmethod
    def collate_fn(cls, samples_list: List[Data]) -> List[Batch]:
        r"""
        Creates a Batch object for each sample in the data list.

        Args:
            samples_list (`List[Data]`): the list of graphs to batch

        Returns:
            a list of Batch objects
        """
        return [Batch.from_data_list([d]) for d in samples_list]

    def _get_loader(
        self, indices: list, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Instantiates the data loader.
        Only works with torch.utils.data.DataLoader.

        Args:
            indices (sequence): Indices in the whole set selected for subset
            kwargs (dict): a dictionary of additional arguments to be
                passed to the dataset being loaded.
                Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object
        """
        shuffle = kwargs.pop("shuffle", False)
        assert not shuffle, (
            "Shuffle cannot be True with a single graph " "sequence dataset."
        )
        batch_size = kwargs.pop("batch_size", 1)

        dataset: TemporalDatasetInterface = self._get_dataset(**kwargs)
        dataset = Subset(dataset, indices)

        assert (
            self.exp_seed is not None
        ), "DataLoader seed has not been specified! Is this a bug?"

        # no need to set worker seed in map-stye dataset, see pytorch doc about reproducibility
        # this would also cause transforms based on random sampling to behave in the same way every time
        # kwargs["worker_init_fn"] = lambda worker_id: seed_worker(
        #     worker_id, self.exp_seed
        # )

        # Using Pytorch default DataLoader instead of PyG, to return list of
        # graphs
        assert (
            self.data_loader_class == torch.utils.data.DataLoader
        ), "SingleGraphSequenceDataProvider accepts a torch DataLoader only!"
        dataloader = self.data_loader_class(
            dataset,
            collate_fn=SingleGraphSequenceDataProvider.collate_fn,
            shuffle=shuffle,
            batch_size=batch_size,
            **kwargs,
        )

        kwargs.update(self.data_loader_args)
        return dataloader


class SingleGraphDataProvider(DataProvider):
    r"""
    A DataProvider subclass that only works with
    :class:`pydgn.data.splitter.SingleGraphSplitter`.

    Args:
        data_root (str): the path of the root folder in which data is stored
        splits_filepath (str): the filepath of the splits. with additional
            metadata
        dataset_class
            (Callable[...,:class:`pydgn.data.dataset.DatasetInterface`]):
            the class of the dataset
        data_loader_class
            (Union[Callable[...,:class:`torch.utils.data.DataLoader`],
            Callable[...,:class:`torch_geometric.loader.DataLoader`]]):
            the class of the data loader to use
        data_loader_args (dict): the arguments of the data loader
        dataset_name (str): the name of the dataset
        outer_folds (int): the number of outer folds for risk assessment.
            1 means hold-out, >1 means k-fold
        inner_folds (int): the number of outer folds for model selection.
            1 means hold-out, >1 means k-fold

    """

    def _get_splitter(self):
        """
        Instantiates the splitter with the parameters stored in the file
        ``self.splits_filepath``. Only works
        with `~pydgn.data.splitter.SingleGraphSplitter`.

        Returns:
            a :class:`~pydgn.data.splitter.Splitter` object
        """
        super()._get_splitter()  # loads splitter into self.splitter
        assert isinstance(
            self.splitter, SingleGraphSplitter
        ), "This class only works with a SingleGraphNodeSplitter splitter."
        return self.splitter

    def _get_dataset(self, **kwargs: dict) -> DatasetInterface:
        """
        Compared to superclass method, this always returns a new instance of
        the dataset, optionally passing extra arguments specified at runtime.

        Args:
            kwargs (dict): a dictionary of additional arguments to be
                passed to the dataset. Not used in the base version

        Returns:
            a :class:`~pydgn.data.dataset.DatasetInterface` object
        """
        assert "shuffle" not in kwargs, (
            "Your implementation of _get_loader should remove `shuffle` "
            "from kwargs before calling _get_dataset"
        )
        assert "batch_size" not in kwargs, (
            "Your implementation of _get_loader should remove `batch_size` "
            "from kwargs before calling _get_dataset"
        )

        # we probably need to pass run-time specific parameters, so load the
        # dataset in memory again
        # an example is the subset of urls in Iterable style datasets
        dataset = load_dataset(
            self.data_root, self.dataset_name, self.dataset_class, **kwargs
        )

        self.dim_node_features = dataset.dim_node_features
        self.dim_edge_features = dataset.dim_edge_features
        self.dim_target = dataset.dim_target

        return dataset

    def _get_loader(
        self, eval_indices: list, training_indices: list, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Compared to superclass method, returns a dataloader with the single
        graph augmented with additional fields. These are `training_indices`
        with the indices that refer to training nodes (usually always
        available) and `eval_indices`, which specify which are the indices
        on which to evaluate (can be validation or test).

        Args:
            indices (sequence): Indices in the whole set selected for subset
            eval_set (bool): whether or not indices refer to eval set
                (validation or test) or to training
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version
        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object
        """
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        dataset: DatasetInterface = self._get_dataset(**kwargs)

        dataset.data.training_indices = torch.tensor(training_indices)
        dataset.data.eval_indices = torch.tensor(eval_indices)

        assert (
            self.exp_seed is not None
        ), "DataLoader's seed has not been specified! Is this a bug?"

        # no need to set worker seed in map-stye dataset, see pytorch doc about reproducibility
        # this would also cause transforms based on random sampling to behave in the same way every time
        # kwargs["worker_init_fn"] = lambda worker_id: seed_worker(
        #     worker_id, self.exp_seed
        # )

        kwargs.update(self.data_loader_args)

        if shuffle is True:
            sampler = RandomSampler(dataset)
            dataloader = self.data_loader_class(
                dataset, sampler=sampler, batch_size=batch_size, **kwargs
            )
        else:
            dataloader = self.data_loader_class(
                dataset, shuffle=False, batch_size=batch_size, **kwargs
            )

        return dataloader

    def get_inner_train(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for model selection associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.inner_folds[self.outer_k][
            self.inner_k
        ].train_idxs
        return self._get_loader(
            train_indices, training_indices=train_indices, **kwargs
        )

    def get_inner_val(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for model selection associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.inner_folds[self.outer_k][
            self.inner_k
        ].train_idxs
        val_indices = splitter.inner_folds[self.outer_k][self.inner_k].val_idxs
        return self._get_loader(
            val_indices, training_indices=train_indices, **kwargs
        )

    def get_outer_train(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the training set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()

        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        return self._get_loader(
            train_indices, training_indices=train_indices, **kwargs
        )

    def get_outer_val(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the validation set for risk assessment associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded.Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        val_indices = splitter.outer_folds[self.outer_k].val_idxs
        return self._get_loader(
            val_indices, training_indices=train_indices, **kwargs
        )

    def get_outer_test(
        self, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        Returns the test set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        test_indices = splitter.outer_folds[self.outer_k].test_idxs
        return self._get_loader(
            test_indices, training_indices=train_indices, **kwargs
        )


class LinkPredictionSingleGraphDataProvider(DataProvider):
    r"""
    An extension of the DataProvider class to deal with link prediction
    on a single graph. Designed to work with
    :class:`~pydgn.data.splitter.LinkPredictionSingleGraphSplitter`.
    We also assume the single-graph dataset can fit in memory
    **WARNING**: this class **modifies** the dataset by creating copies.
    It may not work if a "shared dataset" feature is added to PyDGN.
    """

    def _get_dataset(self, **kwargs):
        """
        Compared to superclass method, this always returns a new instance of
        the dataset, optionally passing extra arguments specified at runtime.

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset. Not used in the base version

        Returns:
            a :class:`~pydgn.data.dataset.DatasetInterface` object
        """
        assert "shuffle" not in kwargs, (
            "Your implementation of _get_loader should remove `shuffle` "
            "from kwargs before calling _get_dataset"
        )
        assert "batch_size" not in kwargs, (
            "Your implementation of _get_loader should remove `batch_size` "
            "from kwargs before calling _get_dataset"
        )

        # Since we modify the dataset, we need different istances of the same
        # graph
        return load_dataset(
            self.data_root, self.dataset_name, self.dataset_class, **kwargs
        )

    def _get_splitter(self):
        """
        Instantiates the splitter with the parameters stored in the file
            ``self.splits_filepath``. Only works with
            `~pydgn.data.splitter.LinkPredictionSingleGraphSplitter`.

        Returns:
            a :class:`~pydgn.data.splitter.Splitter` object
        """
        super()._get_splitter()  # loads splitter into self.splitter
        assert isinstance(self.splitter, LinkPredictionSingleGraphSplitter), (
            "This class only work with a "
            "LinkPredictionSingleGraphSplitter splitter."
        )
        return self.splitter

    def _get_loader(
        self, indices: list, **kwargs: dict
    ) -> Union[torch.utils.data.DataLoader, torch_geometric.loader.DataLoader]:
        r"""
        This method returns a data loader for a **single** graph augmented with
        additional fields. The `y` field becomes (y, positive EVAL edges,
        negative EVAL edges), where eval means these are the edges on which
        to evaluate losses and scores (in fact, eval could also mean
        training!). A list of different Data objects is created, where
        the evaluation edges are randomly permuted. This depends on the size
        of the batch that is specified.

        Args:
            indices (sequence): Indices in the whole set selected for subset
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version
        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object
        """
        # We may want to shuffle the edges of our single graph and take edge
        # batches
        # NOTE: EVAL edges can be TRAINING/VAL/TEST. It is on "eval" edges
        # that we compute the loss (and eventually do training)
        # changing names may be good
        shuffle = kwargs.pop("shuffle", False)
        batch_size = kwargs.pop("batch_size", 1)

        dataset = self._get_dataset(
            **kwargs
        )  # Custom implementation, we need a copy of the dataset every time
        assert len(dataset) == 1, (
            f"Provider accepts a single-graph dataset only, "
            f"but I see {len(dataset)} graphs"
        )
        y = dataset.data.y

        # Use indices to change edge_index and provide a y
        train, eval = indices
        pos_train_edges, train_attr, neg_train_edges = train
        pos_train_edges = torch.tensor(pos_train_edges, dtype=torch.long)
        train_attr = (
            train_attr.clone() if train_attr is not None else None
        )
        neg_train_edges = torch.tensor(neg_train_edges, dtype=torch.long)

        pos_eval_edges, eval_attr, neg_eval_edges = eval
        pos_eval_edges = torch.tensor(pos_eval_edges, dtype=torch.long)
        eval_attr = eval_attr.clone() if eval_attr is not None else None
        neg_eval_edges = torch.tensor(neg_eval_edges, dtype=torch.long)

        # The code below works because we are working with a single graph!
        dataset.data.edge_index = pos_train_edges
        dataset.data.edge_attr = train_attr
        # Eval can be training/val/test
        dataset.data.y = (y, pos_eval_edges, neg_eval_edges)

        if shuffle is True:
            pos_edge_indices = torch.randperm(pos_eval_edges.shape[1])
            neg_edge_indices = torch.randperm(neg_eval_edges.shape[1])
        else:
            pos_edge_indices = torch.arange(pos_eval_edges.shape[1])
            neg_edge_indices = torch.arange(neg_eval_edges.shape[1])

        # create batch of edges here, leaving other fields as is
        batched_edge_dataset = []

        permuted_pos_eval_edges = pos_eval_edges[:, pos_edge_indices]
        permuted_neg_eval_edges = neg_eval_edges[:, neg_edge_indices]
        permuted_edge_attr = (
            eval_attr[pos_edge_indices] if eval_attr is not None else None
        )
        done = False
        batch_start = 0
        while not done:
            if batch_size == 0:
                # Full batch
                pos_batch_end = permuted_pos_eval_edges.shape[1]
                neg_batch_end = permuted_neg_eval_edges.shape[1]
            else:
                # Create subgraph
                pos_batch_end = (
                    batch_start + batch_size
                    if (batch_start + batch_size)
                    < permuted_pos_eval_edges.shape[1]
                    else permuted_pos_eval_edges.shape[1]
                )
                neg_batch_end = (
                    batch_start + batch_size
                    if (batch_start + batch_size)
                    < permuted_neg_eval_edges.shape[1]
                    else permuted_neg_eval_edges.shape[1]
                )

            batch_pos_edge_indices = permuted_pos_eval_edges[
                :, batch_start:pos_batch_end
            ]
            batch_neg_edge_indices = permuted_neg_eval_edges[
                :, batch_start:neg_batch_end
            ]
            batch_edge_attr = (
                permuted_edge_attr[batch_start:pos_batch_end]
                if eval_attr is not None
                else None
            )
            batch_start += batch_size

            if (
                pos_batch_end == permuted_pos_eval_edges.shape[1]
                and neg_batch_end == permuted_neg_eval_edges.shape[1]
            ):
                done = True

            # create data object and append to list
            subgraph = Data(
                x=dataset.data.x,
                edge_index=dataset.data.edge_index,
                edge_attr=dataset.data.edge_attr,
                y=(y, batch_pos_edge_indices, batch_neg_edge_indices),
            )
            batched_edge_dataset.append(subgraph)

        assert (
            self.exp_seed is not None
        ), "DataLoader seed has not been specified! Is this a bug?"

        # no need to set worker seed in map-stye dataset, see pytorch doc about reproducibility
        # this would also cause transforms based on random sampling to behave in the same way every time
        # kwargs["worker_init_fn"] = lambda worker_id: seed_worker(
        #     worker_id, self.exp_seed
        # )

        kwargs.update(self.data_loader_args)

        # Single graph dataset, shuffle does not make sense (unless we know how
        # to do mini-batch training with nodes)
        dataloader = self.data_loader_class(
            batched_edge_dataset, batch_size=1, shuffle=False, **kwargs
        )

        return dataloader

    def get_inner_train(self, **kwargs):
        r"""
        Returns the training set for model selection associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.inner_folds[self.outer_k][
            self.inner_k
        ].train_idxs
        eval_indices = train_indices
        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)

    def get_inner_val(self, **kwargs):
        r"""
        Returns the validation set for model selection associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None and self.inner_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.inner_folds[self.outer_k][
            self.inner_k
        ].train_idxs
        eval_indices = splitter.inner_folds[self.outer_k][
            self.inner_k
        ].val_idxs
        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)

    def get_outer_train(self, **kwargs):
        r"""
        Returns the training set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()

        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        eval_indices = train_indices
        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)

    def get_outer_val(self, **kwargs):
        r"""
        Returns the validation set for risk assessment associated with
        specific outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()

        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        eval_indices = splitter.outer_folds[self.outer_k].val_idxs

        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)

    def get_outer_test(self, **kwargs):
        r"""
        Returns the test set for risk assessment associated with specific
        outer and inner folds

        Args:
            kwargs (dict): a dictionary of additional arguments to be passed
                to the dataset being loaded. Not used in the base version

        Returns:
            a Union[:class:`torch.utils.data.DataLoader`,
            :class:`torch_geometric.loader.DataLoader`] object

        """
        assert self.outer_k is not None
        splitter = self._get_splitter()
        train_indices = splitter.outer_folds[self.outer_k].train_idxs
        eval_indices = splitter.outer_folds[self.outer_k].test_idxs
        indices = train_indices, eval_indices
        return self._get_loader(indices, **kwargs)



import random
from typing import Tuple

import numpy as np
import torch

from sklearn.model_selection import (
    StratifiedKFold,
    StratifiedShuffleSplit,
    ShuffleSplit,
    KFold,
)
from torch_geometric.utils import (
    negative_sampling,
    to_undirected,
    to_dense_adj,
    add_self_loops,
    is_undirected,
)

import pydgn
from pydgn.data.dataset import OGBGDatasetInterface, IterableDatasetInterface
from pydgn.data.util import to_lower_triangular
from pydgn.experiment.util import s2c


class Fold:
    r"""
    Simple class that stores training, validation, and test indices.

    Args:
        train_idxs (Union[list, tuple]): training indices
        val_idxs (Union[list, tuple]): validation indices. Default is ``None``
        test_idxs (Union[list, tuple]): test indices. Default is ``None``
    """

    def __init__(self, train_idxs, val_idxs=None, test_idxs=None):
        self.train_idxs = train_idxs
        self.val_idxs = val_idxs
        self.test_idxs = test_idxs


class InnerFold(Fold):
    r"""
    Simple extension of the Fold class that returns a dictionary with
    training and validation indices (model selection).
    """

    def todict(self) -> dict:
        r"""
        Creates a dictionary with the training/validation indices.

        Returns:
            a dict with keys ``['train', 'val']`` associated with the
            respective indices
        """
        return {"train": self.train_idxs, "val": self.val_idxs}


class OuterFold(Fold):
    r"""
    Simple extension of the Fold class that returns a dictionary with training
    and test indices (risk assessment)
    """

    def todict(self) -> dict:
        r"""
        Creates a dictionary with the training/validation/test indices.

        Returns:
            a dict with keys ``['train', 'val', 'test']``
            associated with the respective indices
        """
        return {
            "train": self.train_idxs,
            "val": self.val_idxs,
            "test": self.test_idxs,
        }


class NoShuffleTrainTestSplit:
    r"""
    Class that implements a very simple training/test split. Can be used to
    further split training data into training and validation.

    Args:
        test_ratio: percentage of data to use for evaluation.
    """

    def __init__(self, test_ratio):
        self.test_ratio = test_ratio

    # Leave the arguments as they are. The parameter `y` is needed to
    # implement the same interface as sklearn
    def split(self, idxs, y=None):
        """
        Splits the data.

        Args:
            idxs: the indices to split according to the `test_ratio` parameter
            y: Unused argument

        Returns:
            a list of a single tuple (train indices, test/eval indices)
        """
        n_samples = len(idxs)
        n_test = int(n_samples * self.test_ratio)
        n_train = n_samples - n_test
        train_idxs = np.arange(n_train)
        test_idxs = np.arange(n_train, n_train + n_test)
        return [(train_idxs, test_idxs)]


class Splitter:
    r"""
    Class that generates and stores the data splits at dataset creation time.

    Args:
        n_outer_folds (int): number of outer folds (risk assessment).
            1 means hold-out, >1 means k-fold
        n_inner_folds (int): number of inner folds (model selection).
            1 means hold-out, >1 means k-fold
        seed (int): random seed for reproducibility (on the same machine)
        stratify (bool): whether to apply stratification or not
            (should be true for classification tasks)
        shuffle (bool): whether to apply shuffle or not
        inner_val_ratio  (float): percentage of validation set for
            hold_out model selection. Default is ``0.1``
        outer_val_ratio  (float): percentage of validation set for
            hold_out model assessment (final training runs).
            Default is ``0.1``
        test_ratio  (float): percentage of test set for
            hold_out model assessment. Default is ``0.1``
    """

    def __init__(
        self,
        n_outer_folds: int,
        n_inner_folds: int,
        seed: int,
        stratify: bool,
        shuffle: bool,
        inner_val_ratio: float = 0.1,
        outer_val_ratio: float = 0.1,
        test_ratio: float = 0.1,
    ):
        self.outer_folds = []
        self.inner_folds = []
        self.stratify = stratify
        self.shuffle = shuffle

        self.n_outer_folds = n_outer_folds
        self.n_inner_folds = n_inner_folds
        self.seed = seed

        self.inner_val_ratio = inner_val_ratio
        self.outer_val_ratio = outer_val_ratio
        self.test_ratio = test_ratio

    def get_targets(
        self, dataset: pydgn.data.dataset.DatasetInterface
    ) -> Tuple[bool, np.ndarray]:
        r"""
        Reads the entire dataset and returns the targets.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`):
                the dataset

        Returns:
            a tuple of two elements. The first element is a boolean, which
            is ``True`` if target values exist or an exception has not
            been thrown. The second value holds the actual targets or ``None``,
            depending on the first boolean value.
        """
        try:
            _ = hasattr(dataset[0], "y")
        except NotImplementedError as e:
            assert issubclass(dataset.__class__, IterableDatasetInterface)
            return False, False

        if hasattr(dataset[0], "y") and dataset[0].y is not None:
            targets = torch.cat([d.y for d in dataset], dim=0).numpy()
            return True, targets
        else:
            return False, None

    @classmethod
    def load(cls, path: str):
        r"""
        Loads the data splits from disk.

        Args:
            :param path: the path of the yaml file with the splits

        Returns:
            a :class:`~pydgn.data.splitter.Splitter` object
        """
        splits = torch.load(path)

        splitter_classname = splits.get("splitter_class", "Splitter")
        splitter_class = s2c(splitter_classname)

        splitter_args = splits.get("splitter_args")
        splitter = splitter_class(**splitter_args)

        assert splitter.n_outer_folds == len(splits["outer_folds"])
        assert splitter.n_inner_folds == len(splits["inner_folds"][0])

        for fold_data in splits["outer_folds"]:
            splitter.outer_folds.append(
                OuterFold(
                    fold_data["train"],
                    val_idxs=fold_data["val"],
                    test_idxs=fold_data["test"],
                )
            )

        for inner_split in splits["inner_folds"]:
            inner_split_data = []
            for fold_data in inner_split:
                inner_split_data.append(
                    InnerFold(fold_data["train"], val_idxs=fold_data["val"])
                )
            splitter.inner_folds.append(inner_split_data)

        return splitter

    def _get_splitter(
        self, n_splits: int, stratified: bool, eval_ratio: float
    ):
        r"""
        Instantiates the appropriate splitter to use depending on the situation

        Args:
            n_splits (int): the number of different splits to create
            stratified (bool): whether or not to perform stratification.
                **Works with graph classification tasks only!**
            eval_ratio (float): the amount of evaluation (validation/test)
                data to use in case ``n_splits==1`` (i.e., hold-out data split)

        Returns:
            a :class:`~pydgn.data.splitter.Splitter` object
        """
        if n_splits == 1:
            if not self.shuffle:
                assert (
                    stratified is False
                ), "Stratified not implemented when shuffle is False"
                splitter = NoShuffleTrainTestSplit(test_ratio=eval_ratio)
            else:
                if stratified:
                    splitter = StratifiedShuffleSplit(
                        n_splits, test_size=eval_ratio, random_state=self.seed
                    )
                else:
                    splitter = ShuffleSplit(
                        n_splits, test_size=eval_ratio, random_state=self.seed
                    )
        elif n_splits > 1:
            if stratified:
                splitter = StratifiedKFold(
                    n_splits, shuffle=self.shuffle, random_state=self.seed
                )
            else:
                splitter = KFold(
                    n_splits, shuffle=self.shuffle, random_state=self.seed
                )
        else:
            raise ValueError(f"'n_splits' must be >=1, got {n_splits}")

        return splitter

    def split(
        self,
        dataset: pydgn.data.dataset.DatasetInterface,
        targets: np.ndarray = None,
    ):
        r"""
        Computes the splits and stores them in the list fields
        ``self.outer_folds`` and ``self.inner_folds``.
        IMPORTANT: calling split() sets the seed of numpy, torch, and
        random for reproducibility.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`):
                the Dataset object
            targets (np.ndarray]): targets used for stratification.
                Default is ``None``
        """
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)
        torch.cuda.manual_seed(self.seed)
        random.seed(self.seed)

        idxs = range(len(dataset))

        stratified = self.stratify
        outer_idxs = np.array(idxs)

        outer_splitter = self._get_splitter(
            n_splits=self.n_outer_folds,
            stratified=stratified,
            eval_ratio=self.test_ratio,
        )  # This is the true test (outer test)

        for train_idxs, test_idxs in outer_splitter.split(
            outer_idxs, y=targets if stratified else None
        ):

            assert set(train_idxs) == set(outer_idxs[train_idxs])
            assert set(test_idxs) == set(outer_idxs[test_idxs])

            inner_fold_splits = []
            inner_idxs = outer_idxs[
                train_idxs
            ]  # equals train_idxs because outer_idxs was ordered
            inner_targets = (
                targets[train_idxs] if targets is not None else None
            )

            inner_splitter = self._get_splitter(
                n_splits=self.n_inner_folds,
                stratified=stratified,
                eval_ratio=self.inner_val_ratio,
            )  # The inner "test" is, instead, the validation set

            for inner_train_idxs, inner_val_idxs in inner_splitter.split(
                inner_idxs, y=inner_targets if stratified else None
            ):
                inner_fold = InnerFold(
                    train_idxs=inner_idxs[inner_train_idxs].tolist(),
                    val_idxs=inner_idxs[inner_val_idxs].tolist(),
                )
                inner_fold_splits.append(inner_fold)

                # False if empty
                assert not bool(
                    set(inner_train_idxs)
                    & set(inner_val_idxs)
                    & set(test_idxs)
                )
                assert not bool(
                    set(inner_idxs[inner_train_idxs])
                    & set(inner_idxs[inner_val_idxs])
                    & set(test_idxs)
                )

            self.inner_folds.append(inner_fold_splits)

            # Obtain outer val from outer train in an holdout fashion
            outer_val_splitter = self._get_splitter(
                n_splits=1,
                stratified=stratified,
                eval_ratio=self.outer_val_ratio,
            )
            outer_train_idxs, outer_val_idxs = list(
                outer_val_splitter.split(inner_idxs, y=inner_targets)
            )[0]

            # False if empty
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(inner_idxs[outer_train_idxs])
                & set(inner_idxs[outer_val_idxs])
                & set(test_idxs)
            )

            np.random.shuffle(outer_train_idxs)
            np.random.shuffle(outer_val_idxs)
            np.random.shuffle(test_idxs)
            outer_fold = OuterFold(
                train_idxs=inner_idxs[outer_train_idxs].tolist(),
                val_idxs=inner_idxs[outer_val_idxs].tolist(),
                test_idxs=outer_idxs[test_idxs].tolist(),
            )
            self.outer_folds.append(outer_fold)

    def _splitter_args(self) -> dict:
        r"""
        Returns a dict with all the splitter's arguments for subsequent
        re-loading at experiment time.

        Returns:
            a dict containing all splitter's arguments.
        """
        return {
            "n_outer_folds": self.n_outer_folds,
            "n_inner_folds": self.n_inner_folds,
            "seed": self.seed,
            "stratify": self.stratify,
            "shuffle": self.shuffle,
            "inner_val_ratio": self.inner_val_ratio,
            "outer_val_ratio": self.outer_val_ratio,
            "test_ratio": self.test_ratio,
        }

    def save(self, path: str):
        r"""
        Saves the split as a dictionary into a ``torch`` file. The arguments
        of the dictionary are
        * seed (int)
        * splitter_class (str)
        * splitter_args (dict)
        * outer_folds (list of dicts)
        * inner_folds (list of lists of dicts)

        Args:
            path (str): filepath where to save the object
        """
        print("Saving splits on disk...")

        # save split class name
        module = self.__class__.__module__
        splitter_class = self.__class__.__qualname__
        if module is not None and module != "__builtin__":
            splitter_class = module + "." + splitter_class

        savedict = {
            "seed": self.seed,
            "splitter_class": splitter_class,
            "splitter_args": self._splitter_args(),
            "outer_folds": [o.todict() for o in self.outer_folds],
            "inner_folds": [],
        }

        for (
            inner_split
        ) in self.inner_folds:  # len(self.inner_folds) == # of **outer** folds
            savedict["inner_folds"].append([i.todict() for i in inner_split])
        torch.save(savedict, path)
        print("Done.")


class SameInnerSplitSplitter(Splitter):
    r"""
    Splitter subclass that can be used to have multiple training runs of the
    same configuration at model selection time. It is not meant to be combined
    with a double-nested CV, for which the different inner splits are already
    enough to gauge the training stability of each configuration.
    """
    def split(
        self,
        dataset: pydgn.data.dataset.DatasetInterface,
        targets: np.ndarray = None,
    ):
        r"""
        Computes the splits and stores them in the list fields
        ``self.outer_folds`` and ``self.inner_folds``.
        IMPORTANT: calling split() sets the seed of numpy, torch, and
        random for reproducibility.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`):
                the Dataset object
            targets (np.ndarray]): targets used for stratification.
                Default is ``None``
        """
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)
        torch.cuda.manual_seed(self.seed)
        random.seed(self.seed)

        idxs = range(len(dataset))

        stratified = self.stratify
        outer_idxs = np.array(idxs)

        outer_splitter = self._get_splitter(
            n_splits=self.n_outer_folds,
            stratified=stratified,
            eval_ratio=self.test_ratio,
        )  # This is the true test (outer test)

        for train_idxs, test_idxs in outer_splitter.split(
            outer_idxs, y=targets if stratified else None
        ):

            assert set(train_idxs) == set(outer_idxs[train_idxs])
            assert set(test_idxs) == set(outer_idxs[test_idxs])

            inner_fold_splits = []
            inner_idxs = outer_idxs[
                train_idxs
            ]  # equals train_idxs because outer_idxs was ordered
            inner_targets = (
                targets[train_idxs] if targets is not None else None
            )

            inner_splitter = self._get_splitter(
                n_splits=self.n_inner_folds,
                stratified=stratified,
                eval_ratio=self.inner_val_ratio,
            )  # The inner "test" is, instead, the validation set

            for inner_train_idxs, inner_val_idxs in inner_splitter.split(
                inner_idxs, y=inner_targets if stratified else None
            ):
                inner_fold = InnerFold(
                    train_idxs=inner_idxs[inner_train_idxs].tolist(),
                    val_idxs=inner_idxs[inner_val_idxs].tolist(),
                )

                # False if empty
                assert not bool(
                    set(inner_train_idxs)
                    & set(inner_val_idxs)
                    & set(test_idxs)
                )
                assert not bool(
                    set(inner_idxs[inner_train_idxs])
                    & set(inner_idxs[inner_val_idxs])
                    & set(test_idxs)
                )

                # we ignore the different inner splits and use only the first
                # one to be reused multiple times (effectively simulating
                # multiple training runs of the same configuration on the same
                # training/validation data split
                for _ in range(self.n_inner_folds):
                    inner_fold_splits.append(inner_fold)
                break

            self.inner_folds.append(inner_fold_splits)

            # Obtain outer val from outer train in an holdout fashion
            outer_val_splitter = self._get_splitter(
                n_splits=1,
                stratified=stratified,
                eval_ratio=self.outer_val_ratio,
            )
            outer_train_idxs, outer_val_idxs = list(
                outer_val_splitter.split(inner_idxs, y=inner_targets)
            )[0]

            # False if empty
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(inner_idxs[outer_train_idxs])
                & set(inner_idxs[outer_val_idxs])
                & set(test_idxs)
            )

            np.random.shuffle(outer_train_idxs)
            np.random.shuffle(outer_val_idxs)
            np.random.shuffle(test_idxs)
            outer_fold = OuterFold(
                train_idxs=inner_idxs[outer_train_idxs].tolist(),
                val_idxs=inner_idxs[outer_val_idxs].tolist(),
                test_idxs=outer_idxs[test_idxs].tolist(),
            )
            self.outer_folds.append(outer_fold)


class TemporalSplitter(Splitter):
    r"""
    Reads the entire dataset and returns the targets. In this case, each
    sample in the dataset represents a temporal graph, so we get the
    classification value at the last time step. Use this method to
    stratify a dataset of multiple temporal graphs.

    Args:
        dataset (:class:`~pydgn.data.dataset.DatasetInterface`): the dataset

    Returns:
        a tuple of two elements. The first element is a boolean, which is
        ``True`` if target values exist or an exception has not been thrown.
        The second value holds the actual targets or ``None``, depending on the
        first boolean value.
    """

    def get_targets(
        self, dataset: pydgn.data.dataset.TemporalDatasetInterface
    ) -> Tuple[bool, np.ndarray]:
        r"""
        Reads the entire dataset and returns the targets.

        Args:
            dataset (:class:`~pydgn.data.dataset.TemporalDatasetInterface`):
                the temporal dataset

        Returns:
            a tuple of two elements. The first element is a boolean, which
            is ``True`` if target values exist or an exception has not
            been thrown. The second value holds the actual targets or
            ``None``, depending on the first boolean value.
        """
        raise NotImplementedError(
            "You should subclass TemporalSplitter and implement this function!"
        )


class OGBGSplitter(Splitter):
    """
    Splitter specific to OGBG Datasets, reuses the already given splits
    (hence it works only in hold-out mode).
    """

    def split(self, dataset: OGBGDatasetInterface, targets=None):
        r"""
        Computes the OGBG splits according to those already provided by
        the authors of the datasets and stores them in the list fields
        ``self.outer_folds`` and ``self.inner_folds``.
        IMPORTANT: calling split() sets the seed of numpy,
        torch, and random for reproducibility.

        Args:
            dataset (:class:`~pydgn.data.dataset.OGBGDatasetInterface`):
                the Dataset object
            targets (np.ndarray]): targets used for stratification.
                Default is ``None``
        """
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)
        torch.cuda.manual_seed(self.seed)
        random.seed(self.seed)

        assert self.n_outer_folds == 1 and self.n_inner_folds == 1, (
            "OGBGSplitter assumes you want to use the same splits "
            "as in the original dataset!"
        )
        original_splits = dataset.get_idx_split()

        outer_train_indices = original_splits["train"].numpy().tolist()
        outer_valid_indices = original_splits["valid"].numpy().tolist()
        outer_test_indices = original_splits["test"].numpy().tolist()

        np.random.shuffle(outer_train_indices)
        np.random.shuffle(outer_valid_indices)
        np.random.shuffle(outer_test_indices)

        inner_fold_splits = []
        inner_fold = InnerFold(
            train_idxs=outer_train_indices, val_idxs=outer_valid_indices
        )
        inner_fold_splits.append(inner_fold)
        self.inner_folds.append(inner_fold_splits)

        outer_fold = OuterFold(
            train_idxs=outer_train_indices,
            val_idxs=outer_valid_indices,
            test_idxs=outer_test_indices,
        )
        self.outer_folds.append(outer_fold)


class SingleGraphSequenceSplitter(TemporalSplitter):
    """
    Class for dynamic graphs that generates the splits at dataset creation
    time. It assumes that there is a single graph sequence, so the split
    happens on time steps. What is more, `n_inner_folds` here will create an
    inner K-fold CV split for model selection where, however, the
    training/validation split will not change (because there is no way to
    split time steps in a different way). This allows for different
    initializations of the same model, evaluating the avg performance
    on the VL set.
    """

    def __init__(
        self,
        n_outer_folds: int,
        n_inner_folds: int,
        seed: int,
        stratify: bool,
        shuffle: bool,
        inner_val_ratio: float = 0.1,
        outer_val_ratio: float = 0.1,
        test_ratio: float = 0.1,
    ):
        super().__init__(
            n_outer_folds,
            n_inner_folds,
            seed,
            stratify,
            shuffle,
            inner_val_ratio,
            outer_val_ratio,
            test_ratio,
        )
        assert n_outer_folds == 1, (
            "SingleGraphSequenceSplitter does not allow more than "
            "one outer fold, as the graph sequence is unique."
        )
        assert not shuffle, (
            "SingleGraphSequenceSplitter does not allow to shuffle "
            "the time steps of the unique graph sequence."
        )
        assert not stratify, "You cannot stratify a single graph sequence."

    def get_targets(
        self, dataset: pydgn.data.dataset.TemporalDatasetInterface
    ) -> Tuple[bool, np.ndarray]:
        r"""
        Reads the entire dataset and returns the targets.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`):
                the dataset

        Returns:
            a tuple of two elements. The first element is a boolean, which is
            ``True`` if target values exist or an exception has not been
            thrown. The second value holds the actual targets or
            ``None``, depending on the first boolean value.
        """
        try:
            targets = np.array([d.targets[-1].item() for d in dataset])
            return True, targets
        except Exception:
            return False, None

    def split(
        self,
        dataset: pydgn.data.dataset.DatasetInterface,
        targets: np.ndarray = None,
    ):
        r"""
        Computes the splits and stores them in the list fields
        ``self.outer_folds`` and ``self.inner_folds``.
        IMPORTANT: calling split() sets the seed of numpy, torch,
        and random for reproducibility.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`):
                the Dataset object
            targets (np.ndarray]): targets used for stratification.
                Default is ``None``
        """
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)
        torch.cuda.manual_seed(self.seed)
        random.seed(self.seed)

        idxs = range(len(dataset))
        print(f"Number of snapshots: {len(dataset)}")

        outer_splitter = self._get_splitter(
            n_splits=self.n_outer_folds,
            stratified=self.stratify,
            eval_ratio=self.test_ratio,
        )  # This is the true test (outer test)

        outer_idxs = np.array(idxs)
        train_idxs, test_idxs = outer_splitter.split(outer_idxs, y=targets)[0]

        assert set(train_idxs) == set(outer_idxs[train_idxs])
        assert set(test_idxs) == set(outer_idxs[test_idxs])

        inner_fold_splits = []
        inner_idxs = outer_idxs[
            train_idxs
        ]  # equals train_idxs because outer_idxs was ordered

        inner_splitter = NoShuffleTrainTestSplit(
            test_ratio=self.inner_val_ratio
        )

        # Repeat the very same inner split for n_inner_folds, as there is no
        # way to split differently.
        # This allows for different initializations of the same model,
        # evaluating the avg performance on the VL set.
        inner_train_idxs, inner_val_idxs = inner_splitter.split(
            inner_idxs, y=None
        )[0]
        for _ in range(self.n_inner_folds):
            inner_fold = InnerFold(
                train_idxs=inner_idxs[inner_train_idxs].tolist(),
                val_idxs=inner_idxs[inner_val_idxs].tolist(),
            )
            inner_fold_splits.append(inner_fold)

        self.inner_folds.append(inner_fold_splits)

        # Obtain outer val from outer train in an holdout fashion
        outer_val_splitter = NoShuffleTrainTestSplit(
            test_ratio=self.outer_val_ratio
        )
        outer_train_idxs, outer_val_idxs = outer_val_splitter.split(
            inner_idxs, y=None
        )[0]

        # False if empty
        assert not bool(
            set(inner_train_idxs) & set(inner_val_idxs) & set(test_idxs)
        )
        assert not bool(
            set(inner_idxs[inner_train_idxs])
            & set(inner_idxs[inner_val_idxs])
            & set(test_idxs)
        )
        assert not bool(
            set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
        )
        assert not bool(
            set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
        )
        assert not bool(
            set(inner_idxs[outer_train_idxs])
            & set(inner_idxs[outer_val_idxs])
            & set(test_idxs)
        )

        outer_fold = OuterFold(
            train_idxs=inner_idxs[outer_train_idxs].tolist(),
            val_idxs=inner_idxs[outer_val_idxs].tolist(),
            test_idxs=outer_idxs[test_idxs].tolist(),
        )
        self.outer_folds.append(outer_fold)


class SingleGraphSplitter(Splitter):
    r"""
    A splitter for a single graph dataset that randomly splits nodes into
    training/validation/test

    Args:
        n_outer_folds (int): number of outer folds (risk assessment).
            1 means hold-out, >1 means k-fold
        n_inner_folds (int): number of inner folds (model selection).
            1 means hold-out, >1 means k-fold
        seed (int): random seed for reproducibility (on the same machine)
        stratify (bool): whether to apply stratification or not
            (should be true for classification tasks)
        shuffle (bool): whether to apply shuffle or not
        inner_val_ratio  (float): percentage of validation set for
            hold_out model selection. Default is ``0.1``
        outer_val_ratio  (float): percentage of validation set for
            hold_out model assessment (final training runs).
            Default is ``0.1``
        test_ratio  (float): percentage of test set for
            hold_out model assessment. Default is ``0.1``
    """

    def split(
        self,
        dataset: pydgn.data.dataset.DatasetInterface,
        targets: np.ndarray = None,
    ):
        r"""
        Compared with the superclass version, the only difference is that the
        range of indices spans across the number of nodes of the single
        graph taken into consideration.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`):
                the Dataset object
            targets (np.ndarray]): targets used for stratification.
                Default is ``None``
        """
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)
        torch.cuda.manual_seed(self.seed)
        random.seed(self.seed)

        assert (
            len(dataset) == 1
        ), "This class works only with single graph datasets"
        idxs = range(dataset.data.x.shape[0])
        stratified = self.stratify
        outer_idxs = np.array(idxs)

        # all this code below could be reused from the original splitter
        outer_splitter = self._get_splitter(
            n_splits=self.n_outer_folds,
            stratified=stratified,
            eval_ratio=self.test_ratio,
        )  # This is the true test (outer test)

        for train_idxs, test_idxs in outer_splitter.split(
            outer_idxs, y=targets
        ):

            assert set(train_idxs) == set(outer_idxs[train_idxs])
            assert set(test_idxs) == set(outer_idxs[test_idxs])

            inner_fold_splits = []
            inner_idxs = outer_idxs[
                train_idxs
            ]  # equals train_idxs because outer_idxs was ordered
            inner_targets = (
                targets[train_idxs] if targets is not None else None
            )

            inner_splitter = self._get_splitter(
                n_splits=self.n_inner_folds,
                stratified=stratified,
                eval_ratio=self.inner_val_ratio,
            )  # The inner "test" is, instead, the validation set

            for inner_train_idxs, inner_val_idxs in inner_splitter.split(
                inner_idxs, y=inner_targets
            ):
                inner_fold = InnerFold(
                    train_idxs=inner_idxs[inner_train_idxs].tolist(),
                    val_idxs=inner_idxs[inner_val_idxs].tolist(),
                )
                inner_fold_splits.append(inner_fold)

                # False if empty
                assert not bool(
                    set(inner_train_idxs)
                    & set(inner_val_idxs)
                    & set(test_idxs)
                )
                assert not bool(
                    set(inner_idxs[inner_train_idxs])
                    & set(inner_idxs[inner_val_idxs])
                    & set(test_idxs)
                )

            self.inner_folds.append(inner_fold_splits)

            # Obtain outer val from outer train in an holdout fashion
            outer_val_splitter = self._get_splitter(
                n_splits=1,
                stratified=stratified,
                eval_ratio=self.outer_val_ratio,
            )
            outer_train_idxs, outer_val_idxs = list(
                outer_val_splitter.split(inner_idxs, y=inner_targets)
            )[0]

            # False if empty
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(outer_train_idxs) & set(outer_val_idxs) & set(test_idxs)
            )
            assert not bool(
                set(inner_idxs[outer_train_idxs])
                & set(inner_idxs[outer_val_idxs])
                & set(test_idxs)
            )

            np.random.shuffle(outer_train_idxs)
            np.random.shuffle(outer_val_idxs)
            np.random.shuffle(test_idxs)
            outer_fold = OuterFold(
                train_idxs=inner_idxs[outer_train_idxs].tolist(),
                val_idxs=inner_idxs[outer_val_idxs].tolist(),
                test_idxs=outer_idxs[test_idxs].tolist(),
            )
            self.outer_folds.append(outer_fold)


class LinkPredictionSingleGraphSplitter(Splitter):
    r"""
    Class that inherits from :class:`~pydgn.data.splitter.Splitter` and
    computes link splits for link classification tasks.
    **IMPORTANT:** This class implements **bootstrapping** rather
    than k-fold cross-validation, so different outer test sets may have
    overlapping indices.

    **Does not support edge attributes at the moment.**

    Args:
        n_outer_folds (int): number of outer folds (risk assessment).
            1 means hold-out, >1 means k-fold
        n_inner_folds (int): number of inner folds (model selection).
            1 means hold-out, >1 means k-fold
        seed (int): random seed for reproducibility (on the same machine)
        stratify (bool): whether to apply stratification or not
            (should be true for classification tasks)
        shuffle (bool): whether to apply shuffle or not
        inner_val_ratio  (float): percentage of validation set for
            hold_out model selection. Default is ``0.1``
        outer_val_ratio  (float): percentage of validation set for
            hold_out model assessment (final training runs).
            Default is ``0.1``
        test_ratio  (float): percentage of test set for
            hold_out model assessment. Default is ``0.1``
        undirected (bool): whether or not the graph is undirected.
            Default is ``False``
        avoid_opposite_negative_edges (bool): whether or not to **avoid**
            creating negative edges that are opposite
            to existing edges Default is ``True``
    """

    def __init__(
        self,
        n_outer_folds: int,
        n_inner_folds: int,
        seed: int,
        stratify: bool,
        shuffle: bool,
        inner_val_ratio: float = 0.1,
        outer_val_ratio: float = 0.1,
        test_ratio: float = 0.1,
        undirected: bool = False,
        avoid_opposite_negative_edges: bool = True,
    ):
        super().__init__(
            n_outer_folds,
            n_inner_folds,
            seed,
            stratify,
            shuffle,
            inner_val_ratio,
            outer_val_ratio,
            test_ratio,
        )
        self.undirected = undirected
        self.avoid_opposite_negative_edges = avoid_opposite_negative_edges

    def train_val_test_edge_split(
        self, edge_index, edge_attr, val_ratio, test_ratio, num_nodes
    ):
        r"""
        Sample training/validation/test edges at random.
        """
        if self.undirected:
            edge_index = to_lower_triangular(edge_index)

        no_edges = edge_index.shape[1]
        permutation = list(range(no_edges))
        random.shuffle(permutation)

        no_val_elements, no_test_elements = int(no_edges * val_ratio), int(
            no_edges * test_ratio
        )

        train_elements = permutation[no_val_elements + no_test_elements :]
        val_elements = permutation[:no_val_elements]
        test_elements = permutation[
            no_val_elements : no_val_elements + no_test_elements
        ]

        assert set(train_elements).intersection(set(test_elements)) == set()
        assert set(train_elements).intersection(set(val_elements)) == set()
        assert set(val_elements).intersection(set(test_elements)) == set()

        train_edges = edge_index[:, train_elements]
        val_edges = edge_index[:, val_elements]
        test_edges = edge_index[:, test_elements]

        if edge_attr is not None:
            train_attr = edge_attr[train_elements]
            val_attr = edge_attr[val_elements]
            test_attr = edge_attr[test_elements]
        else:
            train_attr = None
            val_attr = None
            test_attr = None

        # Restore training edges as undirected for training
        # and double validation and test edges to maintain proportions
        if self.undirected:
            train_edges, train_attr = (
                to_undirected(train_edges, train_attr, num_nodes=num_nodes)
                if train_edges.shape[1] > 0
                else train_edges
            )
            val_edges, val_attr = (
                to_undirected(val_edges, val_attr, num_nodes=num_nodes)
                if val_edges.shape[1] > 0
                else val_edges
            )
            test_edges, test_attr = (
                to_undirected(test_edges, test_attr, num_nodes=num_nodes)
                if test_edges.shape[1] > 0
                else test_edges
            )

        return (
            train_edges,
            train_attr,
            val_edges,
            val_attr,
            test_edges,
            test_attr,
        )

    def split(
        self,
        dataset: pydgn.data.dataset.DatasetInterface,
        targets: np.ndarray = None,
    ):
        r"""
        Computes the splits and stores them in the list fields
        ``self.outer_folds`` and ``self.inner_folds``.
        Links are selected at random: this means outer test folds will
        overlap almost surely with if test_ratio is 10% of the total samples.
        The recommended procedure here is to use the outer folds to do
        bootstrapping rather than k-fold cross-validation. Idea taken
        from: https://arxiv.org/pdf/1811.05868.pdf .
        IMPORTANT: calling split() sets the seed of numpy, torch,
        and random for reproducibility.

        Args:
            dataset (:class:`~pydgn.data.dataset.DatasetInterface`):
                the Dataset object
            targets (np.ndarray]): targets used for stratification.
                Default is ``None``
        """
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)
        torch.cuda.manual_seed(self.seed)
        random.seed(self.seed)

        assert len(dataset) == 1, (
            f"LinkPredictionSingleGraphSplitter works on single graph "
            f"dataset only! Here we have {len(dataset)}"
        )
        edge_index = dataset.data.edge_index
        edge_attr = dataset.data.edge_attr
        num_nodes = dataset.data.x.shape[0]

        if self.undirected:
            assert is_undirected(
                edge_index
            ), "Passed undirected == True with a non-symmetric adj matrix"
        elif not self.undirected:
            assert not is_undirected(
                edge_index
            ), "Passed undirected == False with a non-symmetric adj matrix"

        inner_val_ratio = (
            self.inner_val_ratio
            if self.n_inner_folds == 1
            else float(1 / self.n_inner_folds)
        )
        outer_val_ratio = (
            self.outer_val_ratio
            if self.n_inner_folds == 1
            else float(1 / self.n_inner_folds)
        )
        test_ratio = (
            self.test_ratio
            if self.n_outer_folds == 1
            else float(1 / self.n_outer_folds)
        )

        # Used by negative sampling later
        original_edge_index = edge_index

        # Add self loops to prevent negative edges from being self loops
        original_edge_index_plus_selfloops, _ = add_self_loops(
            original_edge_index
        )

        for outer_k in range(self.n_outer_folds):
            # print(f"Processing splits for outer fold n {outer_k + 1}")

            # Create positive train and test edges for outer fold
            # print(
            #     f"Split positive edges from edge index of shape "
            #     f"{original_edge_index.shape[1]}"
            # )
            (
                pos_train_edges,
                train_attr,
                outer_pos_val_edges,
                outer_val_attr,
                pos_test_edges,
                test_attr,
            ) = self.train_val_test_edge_split(
                original_edge_index,
                edge_attr,
                outer_val_ratio,
                test_ratio,
                num_nodes,
            )

            ###
            # Generate NEGATIVE training/validation/test edges for each
            # fold using the original edge_index
            ###

            # We generate directed negative edges of the same number as true
            # edges
            num_neg_samples = original_edge_index.shape[1]

            # Note: negative edges are directed, but can be considered as
            # undirected if the loss/score is symmetric
            if self.avoid_opposite_negative_edges:
                # We generate symmetric edges to avoid sampling an opposite
                # edge as negative
                # Remember that negative sampling without additional args
                # samples directed edges, so we are good
                negative_edge_index = negative_sampling(
                    to_undirected(original_edge_index_plus_selfloops),
                    num_nodes=num_nodes,
                    num_neg_samples=num_neg_samples,
                )
            else:
                negative_edge_index = negative_sampling(
                    original_edge_index_plus_selfloops,
                    num_nodes=num_nodes,
                    num_neg_samples=num_neg_samples,
                )

            # Negative edges are directed, and it does not make sense to double
            # them.
            tmp_undirected = self.undirected
            self.undirected = False
            (
                neg_train_edges,
                _,
                outer_neg_val_edges,
                _,
                neg_test_edges,
                _,
            ) = self.train_val_test_edge_split(
                negative_edge_index,
                None,
                outer_val_ratio,
                test_ratio,
                num_nodes=num_nodes,
            )
            self.undirected = tmp_undirected

            inner_fold_splits = []
            for inner_k in range(self.n_inner_folds):

                # Do not need to create additional negative edges here! Just
                # split what we already have

                # Create positive train and validation edges for outer fold
                (
                    inner_pos_train_edges,
                    inner_train_attr,
                    inner_pos_val_edges,
                    inner_val_attr,
                    _,
                    _,
                ) = self.train_val_test_edge_split(
                    pos_train_edges,
                    train_attr,
                    inner_val_ratio,
                    0.0,
                    num_nodes=num_nodes,
                )

                # Negative edges are directed, and it does not make sense to
                # double them.
                tmp_undirected = self.undirected
                self.undirected = False
                # Create negative train and validation edges for outer fold
                (
                    inner_neg_train_edges,
                    _,
                    inner_neg_val_edges,
                    _,
                    _,
                    _,
                ) = self.train_val_test_edge_split(
                    neg_train_edges,
                    None,
                    inner_val_ratio,
                    0.0,
                    num_nodes=num_nodes,
                )
                self.undirected = tmp_undirected

                inner_fold = InnerFold(
                    train_idxs=(
                        inner_pos_train_edges.tolist(),
                        inner_train_attr,
                        inner_neg_train_edges.tolist(),
                    ),
                    val_idxs=(
                        inner_pos_val_edges.tolist(),
                        inner_val_attr,
                        inner_neg_val_edges.tolist(),
                    ),
                )
                inner_fold_splits.append(inner_fold)

            self.inner_folds.append(inner_fold_splits)

            outer_fold = OuterFold(
                train_idxs=(
                    pos_train_edges.tolist(),
                    train_attr,
                    neg_train_edges.tolist(),
                ),
                val_idxs=(
                    outer_pos_val_edges.tolist(),
                    outer_val_attr,
                    outer_neg_val_edges.tolist(),
                ),
                test_idxs=(
                    pos_test_edges.tolist(),
                    test_attr,
                    neg_test_edges.tolist(),
                ),
            )
            self.outer_folds.append(outer_fold)

    def _splitter_args(self):
        r"""
        Compared to the superclass version, adds two boolean arguments
        `undirected` and `avoid_opposite_negative_edges`.

        Returns:
            a dict containing all splitter's arguments.
        """
        splitter_args = super()._splitter_args()
        splitter_args.update(
            {
                "undirected": self.undirected,
                "avoid_opposite_negative_edges": self.avoid_opposite_negative_edges,
            }
        )
        return splitter_args



from typing import Tuple, Optional, List

import torch
import torch.nn as nn
from torch_geometric.nn import global_add_pool

from pydgn.model.interface import ReadoutInterface


class LinearGraphReadout(ReadoutInterface):
    """
    Class that implements a simple readout mapping for graph prediction
    """

    def __init__(
        self, dim_node_features, dim_edge_features, dim_target, config
    ):
        super().__init__(
            dim_node_features, dim_edge_features, dim_target, config
        )
        self.W = nn.Linear(dim_node_features, dim_target, bias=True)

    def forward(
        self, node_embeddings: torch.tensor, batch: torch.Tensor, **kwargs
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        """
        Implements a linear graph readout

        Args:
            node_embeddings (`torch.Tensor`): the node embeddings of size `Nxd`
            batch (`torch.Tensor`): a tensor specifying to which graphs
                nodes belong to in the batch
            kwargs (dict): additional parameters (unused)

        Returns:
            a tuple (output, node_embeddings)
        """
        hg = global_add_pool(node_embeddings, batch)
        return self.W(hg), node_embeddings



from typing import List, Optional, Tuple, Union

import torch
from torch import nn
from torch_geometric.data import Batch, Data
from torch_geometric.nn import SAGEConv

# from torch_geometric_temporal import DCRNN

from pydgn.model.interface import ModelInterface


class ToyDGN(ModelInterface):
    """
    A toy Deep Graph Network used to test the library
    """

    def __init__(
        self,
        dim_node_features,
        dim_edge_features,
        dim_target,
        readout_class,
        config,
    ):
        super().__init__(
            dim_node_features,
            dim_edge_features,
            dim_target,
            readout_class,
            config,
        )

        num_layers = config["num_layers"]
        dim_embedding = config["dim_embedding"]
        self.aggregation = config["aggregation"]  # can be mean or max

        if self.aggregation == "max":
            self.fc_max = nn.Linear(dim_embedding, dim_embedding)

        self.readout = readout_class(
            dim_node_features=dim_embedding * num_layers,
            dim_edge_features=dim_edge_features,
            dim_target=dim_target,
            config=config,
        )

        self.layers = nn.ModuleList([])
        for i in range(num_layers):
            dim_input = dim_node_features if i == 0 else dim_embedding

            conv = SAGEConv(dim_input, dim_embedding)
            # Overwrite aggregation method (default is set to mean
            conv.aggr = self.aggregation

            self.layers.append(conv)

    def forward(
        self, data: Batch
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[List[object]]]:
        """
        Implements an Toy DGN with some SAGE graph convolutional layers.

        Args:
            data (torch_geometric.data.Batch): a batch of graphs

        Returns:
            the output depends on the readout passed to the model as argument.
        """
        x, edge_index, batch = data.x, data.edge_index, data.batch

        x_all = []

        for i, layer in enumerate(self.layers):
            x = layer(x, edge_index)
            if self.aggregation == "max":
                x = torch.relu(self.fc_max(x))
            x_all.append(x)

        node_embs = torch.cat(x_all, dim=1)

        return self.readout(node_embs, batch, **dict(edge_index=edge_index))


# class ToyDGNTemporal(ModelInterface):
#     """
#     Simple Temporal Deep Graph Network that can be used to test the library
#     """
#
#     def __init__(
#         self,
#         dim_node_features,
#         dim_edge_features,
#         dim_target,
#         readout_class,
#         config,
#     ):
#         super().__init__(
#             dim_node_features,
#             dim_edge_features,
#             dim_target,
#             readout_class,
#             config,
#         )
#
#         self.dim_embedding = 32
#         filter_size = 1
#
#         self.model = DCRNN(dim_node_features, self.dim_embedding, filter_size)
#         self.linear = nn.Linear(self.dim_embedding, self.dim_embedding)
#
#         self.readout = readout_class(
#             dim_node_features=self.dim_embedding,
#             dim_edge_features=dim_edge_features,
#             dim_target=dim_target,
#             config=config,
#         )
#
#     def forward(self, snapshot: Union[Data, Batch], prev_state=None):
#         """
#         Implements an Toy Temporal DGN with some DCRNN graph
#         convolutional layers.
#
#         Args:
#             snapshot (`Union[Data, Batch]`): a graph or batch of graphs
#                 at timestep t
#             prev_state (`torch.Tensor`): hidden state of the model
#                 (previous time step)
#
#         Returns:
#             the output depends on the readout passed to the model as argument.
#         """
#         # snapshot.x: Tensor of size (num_nodes_t x node_ft_size)
#         # snapshot.edge_index: Adj of size (num_nodes_t x num_nodes_t)
#         x, edge_index, mask = (
#             snapshot.x,
#             snapshot.edge_index,
#             snapshot.time_prediction_mask,
#         )
#
#         h = self.model(x, edge_index, H=prev_state)
#         h = torch.relu(h)
#
#         # Node predictors assume the embedding is in field "x"
#         out, _ = self.readout(h, snapshot.batch)
#
#         return out, h



import os
from pathlib import Path

from torch.utils.tensorboard import SummaryWriter

from pydgn.static import *
from pydgn.training.event.handler import EventHandler
from pydgn.training.event.state import State


class Plotter(EventHandler):
    r"""
    Plotter is the main event handler for plotting at training time.

    Args:
        exp_path (str): path where to store the Tensorboard logs
        kwargs (dict): additional arguments that may depend on the plotter
    """

    def __init__(self, exp_path: str, **kwargs: dict):
        super().__init__()
        self.exp_path = exp_path

        if not os.path.exists(Path(self.exp_path, TENSORBOARD)):
            os.makedirs(Path(self.exp_path, TENSORBOARD))
        self.writer = SummaryWriter(log_dir=Path(self.exp_path, "tensorboard"))

    def on_epoch_end(self, state: State):
        """
        Writes Training, Validation and (if any) Test metrics to Tensorboard

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        for k, v in state.epoch_results[LOSSES].items():
            loss_scalars = {}
            # Remove training/validation/test prefix (coupling with Engine)
            loss_name = " ".join(k.split("_")[1:])
            if TRAINING in k:
                loss_scalars[f"{TRAINING}"] = v
            elif VALIDATION in k:
                loss_scalars[f"{VALIDATION}"] = v
            elif TEST in k:
                loss_scalars[f"{TEST}"] = v

            self.writer.add_scalars(loss_name, loss_scalars, state.epoch)

        for k, v in state.epoch_results[SCORES].items():
            score_scalars = {}
            # Remove training/validation/test prefix (coupling with Engine)
            score_name = " ".join(k.split("_")[1:])
            if TRAINING in k:
                score_scalars[f"{TRAINING}"] = v
            elif VALIDATION in k:
                score_scalars[f"{VALIDATION}"] = v
            elif TEST in k:
                score_scalars[f"{TEST}"] = v

            self.writer.add_scalars(score_name, score_scalars, state.epoch)

    def on_fit_end(self, state: State):
        """
        Frees resources by closing the Tensorboard writer

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        self.writer.close()


class WandbPlotter(EventHandler):
    r"""
    EventHandler subclass for logging to Weights & Biases

    Args:
        wandb_project (str): Project Name for W&B
        wandb_entity (str): Entity Name for W&B
        kwargs (dict): additional arguments that may depend on the plotter
    """

    def __init__(
        self, exp_path: str, wandb_project, wandb_entity, **kwargs: dict
    ):
        super().__init__()
        self.exp_path = exp_path

        try:
            import wandb

            self._wandb = wandb
            self._wandb.require(experiment="service")
            self._wandb.setup()
        except ImportError:
            raise ImportError(
                "To use the Weights and Biases Logger please install wandb."
                "Run `pip install wandb` to install it."
            )

        # Initialize a W&B run
        if self._wandb.run is None:
            self._wandb.init(
                name=self.exp_path, project=wandb_project, entity=wandb_entity
            )

    def on_epoch_end(self, state: State):
        """
        Writes Training, Validation and (if any) Test metrics to WandB

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """

        for k, v in state.epoch_results[LOSSES].items():
            # Remove training/validation/test prefix (coupling with Engine)
            loss_name = " ".join(k.split("_")[1:])
            if TRAINING in k:
                self._wandb.log({f"Train/{loss_name}": v})
            elif VALIDATION in k:
                self._wandb.log({f"Valid/{loss_name}": v})
            elif TEST in k:
                self._wandb.log({f"Test/{loss_name}": v})

        for k, v in state.epoch_results[SCORES].items():
            # Remove training/validation/test prefix (coupling with Engine)
            score_name = " ".join(k.split("_")[1:])
            if TRAINING in k:
                self._wandb.log({f"Train/{score_name}": v})
            elif VALIDATION in k:
                self._wandb.log({f"Valid/{score_name}": v})
            elif TEST in k:
                self._wandb.log({f"Test/{score_name}": v})

    def on_fit_end(self, state: State):
        """
        Frees resources by closing the WandB writer

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        self._wandb.finish()



import copy

from pydgn.experiment.util import s2c
from pydgn.model.interface import ModelInterface
from pydgn.training.event.handler import EventHandler


class Optimizer(EventHandler):
    """
    Optimizer is the main event handler for optimizers.
    Just pass a PyTorch optimizer together with its arguments in the
    configuration file.

    Args:
        model (:class:`~pydgn.model.interface.ModelInterface`):
            the model that has to be trained
        optimizer_class_name (str): dotted path to the optimizer class to use
        accumulate_gradients (bool): if ``True``, accumulate mini-batch
            gradients to perform a batch gradient update without
            loading the entire batch in memory
        kwargs (dict): additional parameters for the specific optimizer
    """

    def __init__(
        self,
        model: ModelInterface,
        optimizer_class_name: str,
        accumulate_gradients: bool = False,
        **kwargs: dict
    ):
        super().__init__()
        self.optimizer = s2c(optimizer_class_name)(
            model.parameters(), **kwargs
        )
        self.accumulate_gradients = accumulate_gradients

    def load_state_dict(self, state_dict):
        """
        Loads the state_dict of the optimizer from a checkpoint

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        self.optimizer.load_state_dict(state_dict)

    def on_fit_start(self, state):
        """
        If a checkpoint is present, load the state of the optimizer

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if state.optimizer_state is not None:
            self.optimizer.load_state_dict(state.optimizer_state)

    def on_training_epoch_start(self, state):
        """
        At the start of epoch, and if the gradient has been accumulated
        across the entire epoch, zeroes the gradient of the optimizer.

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if self.accumulate_gradients:
            self.optimizer.zero_grad()

    def on_training_batch_start(self, state):
        """
        At the start of a batch, if batch updates are in order,
        zeroes the gradient of the optimizer

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if not self.accumulate_gradients:
            self.optimizer.zero_grad()

    def on_training_batch_end(self, state):
        """
        At the end of a batch, if batch updates are in order,
        performs a weight update

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if not self.accumulate_gradients:
            self.optimizer.step()

    def on_training_epoch_end(self, state):
        """
        At the end of a batch, and if the gradient has been
        accumulated across the entire epoch, performs a weight update

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        if self.accumulate_gradients:
            self.optimizer.step()

    def on_epoch_end(self, state):
        """
        Updates the state of the optimizer into the state
        at the end of the epoch

        Args:
            state (:class:`~training.event.state.State`):
                object holding training information
        """
        state.update(
            optimizer_state=copy.deepcopy(self.optimizer.state_dict())
        )



import os

from torch.utils.data import DataLoader
from torch_geometric.data import Batch

from pydgn.data.sampler import RandomSampler
from pydgn.experiment.experiment import Experiment
from pydgn.static import LOSS, SCORE


class SemiSupervisedTask(Experiment):
    """
    Class that implements a semi-supervised experiment.
    There is an ``unsupervised_config`` field in the configuration
    file that is used, together with the field ``model``, to produce
    unsupervised embeddings. These are later used
    by a **readout** extracted from a ``supervised_config``
    field in the configuration file to perform the supervised
    task.
    """

    def __init__(self, model_configuration, exp_path, exp_seed):
        super(SemiSupervisedTask, self).__init__(
            model_configuration, exp_path, exp_seed
        )
        self.root_exp_path = (
            exp_path  # to distinguish unsup. and sup. exp paths
        )

    def run_valid(self, dataset_getter, logger):
        r"""
        This function returns the training and validation results for a
        `model selection run`.
        **Do not attempt to load the test set inside this method!**
        **If possible, rely on already available subclasses of this class**.

        It implements a semi-supervised training scheme where first a model is
        trained (possibly with unsupervised loss,
        hence the name semi-supervised) and a list of Data objects are produced
        according to the training engine used.
        Then a second training is carried out where the data list is used as
        new dataset. The result of this second
        training is used for the model selection procedure.

        Args:
            dataset_getter (:class:`~pydgn.data.provider.DataProvider`):
                a data provider
            logger (:class:`~pydgn.log.logger.Logger`): the logger

        Returns:
            a tuple of training and test dictionaries.
            Each dictionary has two keys:

            * ``LOSS`` (as defined in ``pydgn.static``)
            * ``SCORE`` (as defined in ``pydgn.static``)

            For instance, training_results[SCORE] is a dictionary itself
            with other fields to be used by the evaluator.
        """
        unsupervised_config = self.model_config.unsupervised_config
        supervised_config = self.model_config.supervised_config

        # ------------ PART I: Unsupervised Training ------------ #

        batch_size = unsupervised_config["batch_size"]
        shuffle = (
            unsupervised_config["shuffle"]
            if "shuffle" in unsupervised_config
            else True
        )

        # Instantiate the Dataset
        train_loader = dataset_getter.get_inner_train(
            batch_size=batch_size, shuffle=shuffle
        )
        val_loader = dataset_getter.get_inner_val(
            batch_size=batch_size, shuffle=shuffle
        )

        # Call this after the loaders: the datasets may need to be instantiated
        # with additional parameters
        dim_node_features = dataset_getter.get_dim_node_features()
        dim_edge_features = dataset_getter.get_dim_edge_features()
        dim_target = dataset_getter.get_dim_target()

        # Change exp path to allow Stop & Resume
        self.exp_path = os.path.join(
            self.root_exp_path, f"unsupervised_training"
        )

        # Istantiate the Model
        model = self.create_unsupervised_model(
            dim_node_features, dim_edge_features, dim_target
        )

        # Istantiate the engine (it handles the training loop and the inference
        # phase by abstracting the specifics)
        unsupervised_training_engine = self.create_unsupervised_engine(model)

        (
            _,
            _,
            train_data_list,
            _,
            _,
            val_data_list,
            _,
            _,
            _,
        ) = unsupervised_training_engine.train(
            train_loader=train_loader,
            validation_loader=val_loader,
            test_loader=None,
            max_epochs=unsupervised_config["epochs"],
            logger=logger,
        )

        # ------------ PART II: Supervised Training ------------ #

        # Get the embedding size from the first graph in the dataset
        embedding_size = train_data_list[0].x.shape[1]

        # Get information from the supervised configuration
        batch_size = supervised_config["batch_size"]
        shuffle = (
            supervised_config["shuffle"]
            if "shuffle" in supervised_config
            else True
        )

        def collate_fn(data_list):
            return Batch.from_data_list(data_list)

        # Instantiate the Embedding Dataset for supervised learning
        train_loader = DataLoader(
            train_data_list,
            batch_size,
            sampler=RandomSampler(train_data_list) if shuffle else None,
            shuffle=shuffle if not shuffle else None,
            collate_fn=collate_fn,
        )
        val_loader = DataLoader(
            val_data_list,
            batch_size,
            sampler=RandomSampler(val_data_list) if shuffle else None,
            shuffle=shuffle if not shuffle else None,
            collate_fn=collate_fn,
        )

        # Change exp path to allow Stop & Resume
        self.exp_path = os.path.join(
            self.root_exp_path, f"supervised_training"
        )

        # Instantiate the Model
        model = self.create_supervised_model(
            dim_node_features=embedding_size,
            dim_edge_features=0,
            dim_target=dim_target,
        )

        # Instantiate the engine (it handles the training loop and the
        # inference phase by abstracting the specifics)
        supervised_training_engine = self.create_supervised_engine(model)

        (
            train_loss,
            train_score,
            _,
            val_loss,
            val_score,
            _,
            _,
            _,
            _,
        ) = supervised_training_engine.train(
            train_loader=train_loader,
            validation_loader=val_loader,
            test_loader=None,
            max_epochs=supervised_config["epochs"],
            logger=logger,
        )

        train_res = {LOSS: train_loss, SCORE: train_score}
        val_res = {LOSS: val_loss, SCORE: val_score}
        return train_res, val_res

    def run_test(self, dataset_getter, logger):
        """
        This function returns the training, validation and test results
        for a `final run`.
        **Do not use the test to train the model nor for early stopping!**
        **If possible, rely on already available subclasses of this class**.

        It implements a semi-supervised training scheme where first a model
        is trained (possibly with unsupervised loss,
        hence the name semi-supervised) and a list of Data objects are produced
        according to the training engine used.
        Then a second training is carried out where the data list is used as
        new dataset. The result of this second
        training is used for the risk assessment procedure.

        Args:
            dataset_getter (:class:`~pydgn.data.provider.DataProvider`):
                a data provider
            logger (:class:`~pydgn.log.logger.Logger`): the logger

        Returns:
            a tuple of training,validation,test dictionaries.
            Each dictionary has two keys:

            * ``LOSS`` (as defined in ``pydgn.static``)
            * ``SCORE`` (as defined in ``pydgn.static``)

            For instance, training_results[SCORE] is a dictionary itself with
            other fields to be used by the evaluator.
        """
        unsupervised_config = self.model_config.unsupervised_config
        supervised_config = self.model_config.supervised_config

        # ------------ PART I: Unsupervised Training ------------ #

        batch_size = unsupervised_config["batch_size"]
        shuffle = (
            unsupervised_config["shuffle"]
            if "shuffle" in unsupervised_config
            else True
        )

        # Instantiate the Dataset
        train_loader = dataset_getter.get_outer_train(
            batch_size=batch_size, shuffle=shuffle
        )
        val_loader = dataset_getter.get_outer_val(
            batch_size=batch_size, shuffle=shuffle
        )
        test_loader = dataset_getter.get_outer_test(
            batch_size=batch_size, shuffle=shuffle
        )

        # Call this after the loaders: the datasets may need to be instantiated
        # with additional parameters
        dim_node_features = dataset_getter.get_dim_node_features()
        dim_edge_features = dataset_getter.get_dim_edge_features()
        dim_target = dataset_getter.get_dim_target()

        # Change exp path to allow Stop & Resume
        self.exp_path = os.path.join(
            self.root_exp_path, f"unsupervised_training"
        )

        # Instantiate the Model
        model = self.create_unsupervised_model(
            dim_node_features, dim_edge_features, dim_target
        )

        # Instantiate the engine (it handles the training loop and the
        # inference phase by abstracting the specifics)
        unsupervised_training_engine = self.create_unsupervised_engine(model)

        (
            _,
            _,
            train_data_list,
            _,
            _,
            val_data_list,
            _,
            _,
            test_data_list,
        ) = unsupervised_training_engine.train(
            train_loader=train_loader,
            validation_loader=val_loader,
            test_loader=test_loader,
            max_epochs=unsupervised_config["epochs"],
            logger=logger,
        )

        # ------------ PART II: Supervised Training ------------ #

        # Get the embedding size from the first graph in the dataset
        embedding_size = train_data_list[0].x.shape[1]

        # Get information from the supervised configuration
        batch_size = supervised_config["batch_size"]
        shuffle = (
            supervised_config["shuffle"]
            if "shuffle" in supervised_config
            else True
        )

        def collate_fn(data_list):
            return Batch.from_data_list(data_list)

        # Instantiate the Embedding Dataset for supervised learning
        train_loader = DataLoader(
            train_data_list,
            batch_size,
            sampler=RandomSampler(train_data_list) if shuffle else None,
            shuffle=shuffle if not shuffle else None,
            collate_fn=collate_fn,
        )
        val_loader = DataLoader(
            val_data_list,
            batch_size,
            sampler=RandomSampler(val_data_list) if shuffle else None,
            shuffle=shuffle if not shuffle else None,
            collate_fn=collate_fn,
        )
        test_loader = DataLoader(
            test_data_list,
            batch_size,
            sampler=RandomSampler(test_data_list) if shuffle else None,
            shuffle=shuffle if not shuffle else None,
            collate_fn=collate_fn,
        )

        # Change exp path to allow Stop & Resume
        self.exp_path = os.path.join(
            self.root_exp_path, f"supervised_training"
        )

        # Instantiate the Model
        model = self.create_supervised_model(
            dim_node_features=embedding_size,
            dim_edge_features=0,
            dim_target=dim_target,
        )

        # Instantiate the engine (it handles the training loop and the
        # inference phase by abstracting the specifics)
        supervised_training_engine = self.create_supervised_engine(model)

        (
            train_loss,
            train_score,
            _,
            val_loss,
            val_score,
            _,
            test_loss,
            test_score,
            _,
        ) = supervised_training_engine.train(
            train_loader=train_loader,
            validation_loader=val_loader,
            test_loader=test_loader,
            max_epochs=supervised_config["epochs"],
            logger=logger,
        )

        train_res = {LOSS: train_loss, SCORE: train_score}
        val_res = {LOSS: val_loss, SCORE: val_score}
        test_res = {LOSS: test_loss, SCORE: test_score}
        return train_res, val_res, test_res



import os

import torch

from pydgn.static import *


def atomic_save(data: dict, filepath: str):
    r"""
    Atomically stores a dictionary that can be serialized by
    :func:`torch.save`, exploiting the atomic :func:`os.replace`.

    Args:
        data (dict): the dictionary to be stored
        filepath (str): the absolute filepath where to store the dictionary
    """
    try:
        tmp_path = str(filepath) + ATOMIC_SAVE_EXTENSION
        torch.save(data, tmp_path)
        os.replace(tmp_path, filepath)
    except Exception as e:
        os.remove(tmp_path)
        raise e
